from __future__ import annotations

import torch
import torch.nn as nn
import triton
import triton.language as tl
import warnings
import time
from typing import Sequence, Optional

# In the latest triton, math functions were shuffled around into different modules:
# https://github.com/openai/triton/pull/3172
if hasattr(tl.extra.cuda, "libdevice"):
    tlmath = tl.extra.cuda.libdevice
else:
    tlmath = tl.math

from pyjuice.nodes import ProdNodes
from pyjuice.utils.parameter_list import FastParamList
from pyjuice.utils.kernel_launcher import triton_jit
from .layer import Layer
from .backend.node_partition import partition_nodes_by_n_edges
from .backend.index_set import batched_index_set, batched_index_cum
from .compilation import next_power_of_2, get_prod_layer_stats, prod_layer_forward_compilation, \
                         flatten_c_ids, get_prod_layer_parstats, prod_layer_backward_compilation


class ProdLayer(Layer, nn.Module):

    def __init__(self, nodes: Sequence[ProdNodes], global_nid_start: Optional[int] = None, 
                 layer_sparsity_tol: Optional[float] = None, max_num_partitions: Optional[int] = None, 
                 disable_gpu_compilation: bool = False, force_gpu_compilation: bool = False) -> None:
        Layer.__init__(self, nodes)
        nn.Module.__init__(self)

        assert len(nodes) > 0, "No input node."
        assert len(nodes) == len(set(nodes)), "Input node list contains duplicates."

        use_block_sparse_edges = True
        for nid in range(0, len(nodes)):
            if nodes[nid].is_sparse() or nodes[0].block_size != nodes[nid].block_size:
                use_block_sparse_edges = False
                break
        self.use_block_sparse_edges = use_block_sparse_edges

        self.block_size = nodes[0].block_size if self.use_block_sparse_edges else 1

        if global_nid_start is None:
            global_nid_start = self.block_size

        ## Get layer statistics & prepare for compilation ##

        layer_num_nblocks, layer_num_edges, n_chgs = get_prod_layer_stats(
            self.nodes, self.block_size, global_nid_start = global_nid_start, 
            use_block_sparse_edges = self.use_block_sparse_edges
        )

        self.num_nodes = layer_num_nblocks * self.block_size
        self.num_edges = layer_num_edges
        self._layer_nid_range = (global_nid_start, global_nid_start + self.num_nodes)

        # Find a good strategy to partition the nodes into partitions according to their number of children 
        # to minimize total computation cost
        fw_partition_max_chs = partition_nodes_by_n_edges(
            n_chgs, sparsity_tolerance = layer_sparsity_tol, max_num_partitions = max_num_partitions
        )

        # Since the triton kernels require the maximum number children for each partition to be a power of 2,
        # we postprocess the block sizes
        fw_partition_max_chs = torch.unique(next_power_of_2(fw_partition_max_chs))

        self.num_fw_partitions = len(fw_partition_max_chs) # Number of partitions

        # fw_n_partition_ids:      [num_ns]             stores the partition id for each `ns` in `nodes`
        # fw_n_id_in_partition:    [num_ns]             stores the start index of each `ns` in the corresponding partition
        # fw_num_ngs_in_partition: [num_fw_partitions]  number of node blocks in each partition
        num_ns = len(self.nodes)
        fw_n_partition_ids = torch.zeros([num_ns], dtype = torch.long)
        fw_n_id_in_partition = torch.zeros([num_ns], dtype = torch.long)
        fw_num_ngs_in_partition = torch.zeros([self.num_fw_partitions], dtype = torch.long)

        for ns_id, ns in enumerate(self.nodes):
            partition_id = (ns.num_chs > fw_partition_max_chs).sum().item()

            fw_n_partition_ids[ns_id] = partition_id
            fw_n_id_in_partition[ns_id] = fw_num_ngs_in_partition[partition_id]
            if self.use_block_sparse_edges:
                fw_num_ngs_in_partition[partition_id] += ns.num_node_blocks
            else:
                fw_num_ngs_in_partition[partition_id] += ns.num_nodes

        ## Initialize forward pass ##

        # nids:      List[[partition_size]]                      stores node ids
        # cids:      List[[partition_size, partition_max_n_chs]] stores indices of child nodes
        nids, cids = prod_layer_forward_compilation(
            self.nodes, fw_partition_max_chs, fw_n_partition_ids, fw_n_id_in_partition, fw_num_ngs_in_partition, 
            self.block_size, self.use_block_sparse_edges
        )

        # Store buffers for the forward pass
        self.partitioned_nids = FastParamList([nn.Parameter(tensor, requires_grad = False) for tensor in nids])
        self.partitioned_cids = FastParamList([nn.Parameter(tensor, requires_grad = False) for tensor in cids])

        ## Initialize backward pass ##

        # flat_cids:   flattened version of `cids`
        # flat_cid2id: mapping from every `flat_cids` to its corresponding `nids`
        flat_cids, flat_cid2nid = flatten_c_ids(nids, cids)

        # flat_u_cids:        [num_used_ch_nblocks]    child block ids that have at least one parent
        # par_counts:         [num_used_ch_nblocks]    the number of parents for each child node block
        # Note: the dummy node has been removed from `flat_u_cids` and `par_counts`
        flat_u_cids, par_counts = get_prod_layer_parstats(flat_cids, global_nid_start = global_nid_start)

        # Find a good strategy to partition the child nodes into blocks according to their number of parents 
        # to minimize total computation cost
        bk_partition_max_pars = partition_nodes_by_n_edges(
            par_counts, sparsity_tolerance = layer_sparsity_tol, max_num_partitions = max_num_partitions
        )

        # Since the triton kernels require the maximum number children for each block to be a power of 2,
        # we postprocess the block sizes
        bk_partition_max_pars = torch.unique(next_power_of_2(bk_partition_max_pars))

        self.num_bk_partitions = len(bk_partition_max_pars) # Number of partitions

        # bk_n_partition_ids:     [num_ch_nblocks]       stores the block id for each `ns` in `nodes`
        # bk_n_id_in_partition:   [num_ch_nblocks]       stores the start index of each `ns` in the partition
        # bk_num_ns_in_partition: [num_bk_partitions]    number of node blocks in each partition
        num_ch_nblocks = flat_u_cids.size(0)
        bk_n_partition_ids = torch.zeros([num_ch_nblocks], dtype = torch.long)
        bk_n_id_in_partition = torch.zeros([num_ch_nblocks], dtype = torch.long)
        bk_num_ns_in_partition = torch.zeros([self.num_bk_partitions], dtype = torch.long)

        min_n_pars = 0
        for partition_id, max_n_pars in enumerate(bk_partition_max_pars):
            criterion = (par_counts >= min_n_pars) & (par_counts <= max_n_pars)
            filtered_idxs = torch.where(criterion)[0]
            partition_size = criterion.sum().item()

            bk_n_partition_ids[criterion] = partition_id
            bk_n_id_in_partition[criterion] = torch.arange(partition_size)
            bk_num_ns_in_partition[partition_id] = partition_size

            min_n_pars = max_n_pars + 1

        # u_cids:    List[[partition_ch_size]]                       stores child node block ids
        # parids:    List[[partition_ch_size, partition_max_n_pars]] stores indices of parent node blocks
        u_cids, parids = prod_layer_backward_compilation(
            flat_u_cids, flat_cids, flat_cid2nid, 
            bk_partition_max_pars, bk_n_partition_ids, bk_n_id_in_partition, bk_num_ns_in_partition,
            use_cuda = force_gpu_compilation or (not disable_gpu_compilation and (flat_cids.size(0) > 4000))
        )

        # Store buffers for the backward pass
        self.partitioned_u_cids = FastParamList([nn.Parameter(tensor, requires_grad = False) for tensor in u_cids])
        self.partitioned_parids = FastParamList([nn.Parameter(tensor, requires_grad = False) for tensor in parids])

    def forward(self, node_mars: torch.Tensor, element_mars: torch.Tensor, _for_backward: bool = False, **kwargs) -> None:
        """
        Computes the forward pass of a product layer. If `block_size == 1`, it is equivalent to the following:
        ```
        element_mars[nids] = node_mars[cids].sum(dim = 1)
        ```

        Parameters:
        `node_mars`:    [num_nodes, B]
        `element_mars`: [max_num_els, B]
        """

        if not _for_backward and self.provided("fw_partition_local_ids"):
            # Partial evaluation (for forward pass)
            for partition_id in range(self.num_fw_partitions):
                nids = self.partitioned_nids[partition_id]
                cids = self.partitioned_cids[partition_id]
                local_ids = self.fw_partition_local_ids[partition_id]

                self._forward_backward(
                    element_mars, node_mars, nids, cids, local_ids = local_ids, accum = False
                )

        elif _for_backward and self.provided("bk_fw_partition_local_ids"):
            # Partial evaluation (for backward pass)
            for partition_id in range(self.num_fw_partitions):
                nids = self.partitioned_nids[partition_id]
                cids = self.partitioned_cids[partition_id]
                local_ids = self.bk_fw_partition_local_ids[partition_id]

                self._forward_backward(
                    element_mars, node_mars, nids, cids, local_ids = local_ids, accum = False
                )

        else:
            # Evaluate the whole layer
            for partition_id in range(self.num_fw_partitions):
                nids = self.partitioned_nids[partition_id]
                cids = self.partitioned_cids[partition_id]

                self._forward_backward(
                    element_mars, node_mars, nids, cids, accum = False
                )

        return None

    def backward(self, node_flows: torch.Tensor, element_flows: torch.Tensor, logspace_flows: bool = False, **kwargs) -> None:
        """
        Computes the backward pass of a product layer:
        ```
        node_flows[u_cids] = element_flows[parids].sum(dim = 1)
        ```

        Parameters:
        `node_flows`:    [num_nodes, B]
        `element_flows`: [max_num_els, B]
        """
        
        if self.provided("bk_partition_local_ids"):
            # Partial evaluation
            for partition_id in range(self.num_bk_partitions):
                u_cids = self.partitioned_u_cids[partition_id]
                parids = self.partitioned_parids[partition_id]
                local_ids = self.bk_partition_local_ids[partition_id]

                self._forward_backward(node_flows, element_flows, u_cids, parids, local_ids = local_ids, accum = True,
                                       prop_logsumexp = logspace_flows)
        
        else:
            # Evaluate the whole layer
            for partition_id in range(self.num_bk_partitions):
                u_cids = self.partitioned_u_cids[partition_id]
                parids = self.partitioned_parids[partition_id]

                self._forward_backward(node_flows, element_flows, u_cids, parids, accum = True,
                                       prop_logsumexp = logspace_flows)
        
        return None

    def enable_partial_evaluation(self, fw_scopes: Optional[Sequence[BitSet]] = None, bk_scopes: Optional[Sequence[BitSet]] = None):
        super(ProdLayer, self).enable_partial_evaluation(fw_scopes = fw_scopes, bk_scopes = bk_scopes)

        # For product layers, we need a special forward pass during the backward process of the circuit
        if bk_scopes is not None:
            bk_fw_partition_local_ids = [[] for _ in range(self.num_fw_partitions)]
            for scope in bk_scopes:
                if scope not in self.fw_scope2localids:
                    continue

                for partition_id, ids in enumerate(self.fw_scope2localids[scope]):
                    bk_fw_partition_local_ids[partition_id].append(self.fw_scope2localids[scope][partition_id])

            self.bk_fw_partition_local_ids = [
                torch.cat(ids, dim = 0) if len(ids) > 0 else torch.zeros([0], dtype = torch.long) for ids in bk_fw_partition_local_ids
            ]

    def is_prod(self):
        return True

    @staticmethod
    # @triton.jit
    @triton_jit
    def _forward_backward_kernel_3d(node_vals_ptr, element_vals_ptr, local_ids_ptr, nids_ptr, cids_ptr, tot_n_nodes, tot_n_eles, n_nblocks,
                                    num_edges: tl.constexpr, batch_size, BLOCK_M: tl.constexpr, BLOCK_B: tl.constexpr, 
                                    block_size: tl.constexpr, accum: tl.constexpr, partial_eval: tl.constexpr, prop_logsumexp: tl.constexpr):
        """
        This kernel implements the function with 3d tensors. However, it only work with `triton==2.0.0`.
        """
        
        pid_m = tl.program_id(axis = 0) # ID of size-`BLOCK_M` nodes
        pid_b = tl.program_id(axis = 1) # ID of size-`BLOCK_B` batches

        if block_size >= BLOCK_M:

            # Get inferred node block id from `pid_m`
            nblock_id = pid_m // (block_size // BLOCK_M)
            ntile_id = pid_m % (block_size // BLOCK_M)

            # For partial evaluation
            if partial_eval:
                nblock_id = tl.load(local_ids_ptr + nblock_id)

            # Batch offsets and mask
            offs_batch = tl.arange(0, BLOCK_B) + pid_b * BLOCK_B 
            mask_batch = offs_batch < batch_size

            # Get the block start ids for the children
            # To make the triton compiler happy, we reload every index `BLOCK_M` times
            offs_ne = tl.arange(0, num_edges * BLOCK_M) // BLOCK_M
            offs_ne = tl.view(offs_ne, (BLOCK_M, num_edges))
            offs_egstart = tl.load(cids_ptr + nblock_id * num_edges + offs_ne) # [BLOCK_M, num_edges]

            # Get the edge values from child nodes
            block_nids = tl.arange(0, BLOCK_M) + ntile_id * BLOCK_M
            offs_evals = offs_egstart + block_nids[:,None]
            evals = tl.load(element_vals_ptr + offs_evals[None,:,:] * batch_size + offs_batch[:,None,None], mask = mask_batch[:,None,None])

            if prop_logsumexp:
                # Take the logsumexp of the child nodes' values
                evals_max = tl.max(evals, axis = 2)
                nvals = tl.log(tl.sum(tl.exp(evals - evals_max[:,:,None]), axis = 2)) + evals_max
            else:
                # Take the sum of the child nodes' values
                nvals = tl.sum(evals, axis = 2)

            # Node ids to `node_vals_ptr`
            nblock_start = tl.load(nids_ptr + nblock_id)
            offs_nvals = (nblock_start + block_nids[None,:]) * batch_size + offs_batch[:,None]

            # Accumulate the `node_vals` if required
            if accum:
                node_vals = tl.load(node_vals_ptr + offs_nvals, mask = mask_batch[:,None], other = 0)
                
                if prop_logsumexp:
                    # logaddexp
                    diff = nvals - node_vals
                    nvals = tl.where(
                        diff == 0, 
                        nvals + 0.69314718055994530942, # log(2)
                        tl.where(
                            diff > 0,
                            nvals + tlmath.log1p(tl.exp(-diff)),
                            node_vals + tlmath.log1p(tl.exp(diff))
                        )
                    )
                else:
                    # sum
                    nvals += node_vals

            tl.store(node_vals_ptr + offs_nvals, nvals, mask = mask_batch[:,None])

        else:

            # Node offsets and mask
            offs_node = tl.arange(0, BLOCK_M) + pid_m * BLOCK_M
            mask_node = offs_node < n_nblocks * block_size

            # Inferred block ids
            nblock_ids = offs_node // block_size

            # For partial evaluation
            if partial_eval:
                nblock_ids = tl.load(local_ids_ptr + nblock_ids, mask = mask_node)

            # Batch offsets and mask
            offs_batch = tl.arange(0, BLOCK_B) + pid_b * BLOCK_B 
            mask_batch = offs_batch < batch_size

            # Get the block start ids for the children
            offs_ne = tl.arange(0, num_edges * BLOCK_M) // BLOCK_M
            offs_ne = tl.view(offs_ne, (BLOCK_M, num_edges))
            offs_egstart = tl.load(cids_ptr + nblock_ids[:,None] * num_edges + offs_ne, mask = mask_node[:,None]) # [BLOCK_M, num_edges]

            # Get the edge values from child nodes
            block_nids = (offs_node % block_size)
            offs_evals = offs_egstart + block_nids[:,None]
            evals = tl.load(element_vals_ptr + offs_evals[None,:,:] * batch_size + offs_batch[:,None,None], mask = (mask_batch[:,None,None] & mask_node[None,:,None]))

            if prop_logsumexp:
                # Take the logsumexp of the child nodes' values
                evals_max = tl.max(evals, axis = 2)
                nvals = tl.log(tl.sum(tl.exp(evals - evals_max[:,:,None]), axis = 2)) + evals_max
            else:
                # Take the sum of the child nodes' values
                nvals = tl.sum(evals, axis = 2)

            # Node ids to `node_vals_ptr`
            nblock_start = tl.load(nids_ptr + nblock_ids[None,:])
            offs_nvals = (nblock_start + block_nids[None,:]) * batch_size + offs_batch[:,None]

            # Accumulate the `node_vals` if required
            if accum:
                node_vals = tl.load(node_vals_ptr + offs_nvals, mask = mask_batch[:,None], other = 0)
                
                if prop_logsumexp:
                    # logaddexp
                    diff = node_vals - nvals
                    nvals = tl.where(
                        nvals == -float("inf"), 
                        node_vals,
                        tl.where(
                            diff > 0,
                            node_vals + tlmath.log1p(tl.exp(-diff)),
                            nvals + tlmath.log1p(tl.exp(diff))
                        )
                    )
                else:
                    # sum
                    nvals += node_vals

            tl.store(node_vals_ptr + offs_nvals, nvals, mask = mask_batch[:,None])

    @staticmethod
    # @triton.jit
    @triton_jit
    def _forward_backward_kernel_2d(node_vals_ptr, element_vals_ptr, local_ids_ptr, nids_ptr, cids_ptr, tot_n_nodes, tot_n_eles, n_nblocks,
                                    num_edges: tl.constexpr, batch_size, BLOCK_M: tl.constexpr, BLOCK_B: tl.constexpr, 
                                    block_size: tl.constexpr, accum: tl.constexpr, partial_eval: tl.constexpr, prop_logsumexp: tl.constexpr):
        """
        This kernel implements the function with 2d tensors. It works for all `triton` versions.
        """

        pid_m = tl.program_id(axis = 0) # ID of size-`BLOCK_M` nodes
        pid_b = tl.program_id(axis = 1) # ID of size-`BLOCK_B` batches

        # Get inferred node block id from `pid_m`
        nblock_id = pid_m // (block_size // BLOCK_M)
        ntile_id = pid_m % (block_size // BLOCK_M)

        # For partial evaluation
        if partial_eval:
            nblock_id = tl.load(local_ids_ptr + nblock_id)

        # Batch offsets and mask
        offs_batch = tl.arange(0, BLOCK_B) + pid_b * BLOCK_B # [BLOCK_B]
        mask_batch = offs_batch < batch_size

        # Get the block start ids for the children
        offs_edge = tl.arange(0, num_edges)
        offs_egstart = tl.load(cids_ptr + nblock_id * num_edges + offs_edge) # [num_edges]

        # Base ptr for ch values
        offs_evals = (offs_egstart[:,None] + ntile_id * BLOCK_M) * batch_size + offs_batch[None,:] # [num_edges, BLOCK_B]

        # Base ptr for par values
        nblock_start = tl.load(nids_ptr + nblock_id)
        offs_nvals = (nblock_start + ntile_id * BLOCK_M) * batch_size + offs_batch # [BLOCK_B]

        # Inner loop
        for i in range(0, BLOCK_M):
            evals = tl.load(element_vals_ptr + offs_evals, mask = mask_batch[None,:], other = 0)
            
            if prop_logsumexp:
                # Take the logsumexp of the child nodes' values
                evals_max = tl.max(evals, axis = 0)
                nvals = tl.where(evals_max != -float("inf"), tl.log(tl.sum(tl.exp(evals - evals_max[None,:]), axis = 0)) + evals_max, -float("inf"))
            else:
                # Take the sum of the child nodes' values
                nvals = tl.sum(evals, axis = 0)

            # Accumulate the `node_vals` if required
            if accum:
                node_vals = tl.load(node_vals_ptr + offs_nvals, mask = mask_batch)

                if prop_logsumexp:
                    # logaddexp
                    diff = node_vals - nvals
                    nvals = tl.where(
                        nvals == -float("inf"), 
                        node_vals,
                        tl.where(
                            diff > 0,
                            node_vals + tlmath.log1p(tl.exp(-diff)),
                            nvals + tlmath.log1p(tl.exp(diff))
                        )
                    )
                else:
                    # sum
                    nvals += node_vals

            tl.store(node_vals_ptr + offs_nvals, nvals, mask = mask_batch)

            offs_nvals += batch_size
            offs_evals += batch_size

    @staticmethod
    # @triton.jit
    @triton_jit
    def _forward_backward_kernel_large(node_vals_ptr, element_vals_ptr, local_ids_ptr, nids_ptr, cids_ptr, tot_n_nodes, tot_n_eles, n_nblocks,
                                       num_edges: tl.constexpr, batch_size, BLOCK_N: tl.constexpr, BLOCK_B: tl.constexpr, 
                                       N_NUM_BLKS: tl.constexpr, block_size: tl.constexpr, accum: tl.constexpr, partial_eval: tl.constexpr,
                                       prop_logsumexp: tl.constexpr):
        """
        This kernel implements the function with 2d tensors. It is designed for nodes with many edges.
        """
        
        pid_m = tl.program_id(axis = 0) # ID of size-`1` nodes
        pid_b = tl.program_id(axis = 1) # ID of size-`BLOCK_B` batches

        # Get inferred node block id from `pid_m`
        nblock_id = pid_m // block_size
        ntile_id = pid_m % block_size

        # For partial evaluation
        if partial_eval:
            nblock_id = tl.load(local_ids_ptr + nblock_id)

        # Batch offsets and mask
        offs_batch = tl.arange(0, BLOCK_B) + pid_b * BLOCK_B # [BLOCK_B]
        mask_batch = offs_batch < batch_size

        # Get the block start ids for the children
        offs_edge = tl.arange(0, BLOCK_N)
        mask_edge = (offs_edge < num_edges)
        offs_egstart = tl.load(cids_ptr + nblock_id * num_edges + offs_edge) # [BLOCK_N]

        # Base ptr for ch values
        offs_evals = (offs_egstart[:,None] + ntile_id) * batch_size + offs_batch[None,:] # [BLOCK_N, BLOCK_B]

        # Base ptr for par values
        nblock_start = tl.load(nids_ptr + nblock_id)
        offs_nvals = (nblock_start + ntile_id) * batch_size + offs_batch # [BLOCK_B]

        # Prepare buffer
        if prop_logsumexp:
            nvals = tl.zeros([BLOCK_B], dtype = tl.float32) - float("inf")
        else:
            nvals = tl.zeros([BLOCK_B], dtype = tl.float32)

        # Inner loop
        for i in range(0, N_NUM_BLKS):
            evals = tl.load(element_vals_ptr + offs_evals, mask = (mask_edge[:,None] & mask_batch[None,:]), other = 0)
            
            if prop_logsumexp:
                # Take the logsumexp of the child nodes' values
                evals_max = tl.max(evals, axis = 0)
                nvals_sub = tl.where(evals_max != -float("inf"), tl.sum(tl.exp(evals - evals_max[None,:]), axis = 0), 0.0)
                nvals = tl.where(evals_max > nvals,
                    tl.log(nvals_sub + tl.exp(nvals - evals_max) + 1e-24) + evals_max,
                    tl.log(tl.exp(evals_max - nvals) * nvals_sub + 1.0) + nvals
                )
            else:
                # Take the sum of the child nodes' values
                nvals += tl.sum(evals, axis = 0)

            offs_edge += BLOCK_N
            mask_edge = (offs_edge < num_edges)

            # Re-compute the ch value ids
            offs_egstart = tl.load(cids_ptr + nblock_id * num_edges + offs_edge) # [BLOCK_N]
            offs_evals = (offs_egstart[:,None] + ntile_id) * batch_size + offs_batch[None,:] # [BLOCK_N, BLOCK_B]

        # Accumulate the `node_vals` if required
        if accum:
            node_vals = tl.load(node_vals_ptr + offs_nvals, mask = mask_batch)
            
            if prop_logsumexp:
                # logaddexp
                diff = node_vals - nvals
                nvals = tl.where(
                    nvals == -float("inf"), 
                    node_vals,
                    tl.where(
                        diff > 0,
                        node_vals + tlmath.log1p(tl.exp(-diff)),
                        nvals + tlmath.log1p(tl.exp(diff))
                    )
                )
            else:
                # sum
                nvals += node_vals

        tl.store(node_vals_ptr + offs_nvals, nvals, mask = mask_batch)

    def _forward_backward(self, node_vals: torch.Tensor, element_vals: torch.Tensor,
                          nids: torch.Tensor, cids: torch.Tensor, local_ids: Optional[torch.Tensor] = None,
                          accum: bool = False, prop_logsumexp: bool = False) -> None:
        tot_n_nodes = node_vals.size(0)
        tot_n_eles = element_vals.size(0)
        n_nblocks = nids.size(0) if local_ids is None else local_ids.size(0)
        num_edges = cids.size(1)
        batch_size = node_vals.size(1)

        block_size = self.block_size
        partial_eval = local_ids is not None

        assert num_edges & (num_edges - 1) == 0, "`num_edges` must be a power of 2."

        # Special case: every node have > 2048 edges
        if num_edges > 2048:

            BLOCK_N = 2048
            BLOCK_B = 1

            grid = (n_nblocks * self.block_size, batch_size)
            
            self._forward_backward_kernel_large[grid](
                node_vals_ptr = node_vals, 
                element_vals_ptr = element_vals,
                local_ids_ptr = local_ids,
                nids_ptr = nids, 
                cids_ptr = cids, 
                tot_n_nodes = tot_n_nodes,
                tot_n_eles = tot_n_eles,
                n_nblocks = n_nblocks,
                num_edges = num_edges,
                batch_size = batch_size, 
                BLOCK_N = BLOCK_N, 
                BLOCK_B = BLOCK_B, 
                N_NUM_BLKS = triton.cdiv(num_edges, BLOCK_B), 
                block_size = block_size, 
                accum = accum, 
                partial_eval = partial_eval,
                prop_logsumexp = prop_logsumexp
            )

            return None

        if not triton.__version__ == "2.0.0":

            BLOCK_B = min(2048 // num_edges, triton.next_power_of_2(batch_size))
            BLOCK_M = min(max(2048 // (BLOCK_B * num_edges), 1), self.block_size)

            grid = (triton.cdiv(n_nblocks * self.block_size, BLOCK_M), triton.cdiv(batch_size, BLOCK_B))

            self._forward_backward_kernel_2d[grid](
                node_vals_ptr = node_vals, 
                element_vals_ptr = element_vals,
                local_ids_ptr = local_ids,
                nids_ptr = nids, 
                cids_ptr = cids, 
                tot_n_nodes = tot_n_nodes,
                tot_n_eles = tot_n_eles,
                n_nblocks = n_nblocks,
                num_edges = num_edges,
                batch_size = batch_size,
                BLOCK_M = BLOCK_M, 
                BLOCK_B = BLOCK_B,
                block_size = block_size,
                accum = accum,
                partial_eval = partial_eval,
                prop_logsumexp = prop_logsumexp
            )

        else:

            BLOCK_B = min(1024 // num_edges, triton.next_power_of_2(batch_size))
            BLOCK_M = min(max(1024 // (BLOCK_B * num_edges), 1), triton.next_power_of_2(n_nblocks) * self.block_size)

            grid = (triton.cdiv(n_nblocks * self.block_size, BLOCK_M), triton.cdiv(batch_size, BLOCK_B))

            self._forward_backward_kernel_3d[grid](
                node_vals_ptr = node_vals, 
                element_vals_ptr = element_vals,
                local_ids_ptr = local_ids,
                nids_ptr = nids, 
                cids_ptr = cids, 
                tot_n_nodes = tot_n_nodes,
                tot_n_eles = tot_n_eles,
                n_nblocks = n_nblocks,
                num_edges = num_edges,
                batch_size = batch_size,
                BLOCK_M = BLOCK_M, 
                BLOCK_B = BLOCK_B,
                block_size = block_size,
                accum = accum,
                partial_eval = partial_eval,
                prop_logsumexp = prop_logsumexp
            )

        return None

    def _prepare_scope2nids(self):

        # Saved for the next sum layer
        prod_scope_eleids = list()
        global_eid = 1
        for ns in self.nodes:
            s_eid = global_eid
            e_eid = global_eid + ns.num_nodes

            prod_scope_eleids.append((ns.scope, (s_eid, e_eid)))

            global_eid += ns.num_nodes

        if not (hasattr(self, "fw_scope2localids") and hasattr(self, "bk_scope2localids")):
            fw_scope2localids = dict()
            bk_scope2localids = dict()

            # Forward local indices
            global_eid = 1
            for ns in self.nodes:
                scope = ns.scope

                s_eid = global_eid
                e_eid = global_eid + ns.num_nodes

                with torch.no_grad():
                    if scope not in fw_scope2localids:
                        fw_scope2localids[scope] = [
                            torch.zeros([0], dtype = torch.long).to(self.partitioned_nids[0].device) for _ in range(self.num_fw_partitions)
                        ]

                    for partition_id in range(self.num_fw_partitions):
                        nids = self.partitioned_nids[partition_id]
                        partition_local_ids = torch.where((nids >= s_eid) & (nids < e_eid))[0]

                        fw_scope2localids[scope][partition_id] = torch.cat(
                            (fw_scope2localids[scope][partition_id], partition_local_ids), dim = 0
                        )

                global_eid += ns.num_nodes

            # Backward local indices
            for ns in self.nodes:
                for cs in ns.chs:
                    scope = cs.scope

                    s_nid = cs._output_ind_range[0]
                    e_nid = cs._output_ind_range[1]

                    if scope not in bk_scope2localids:
                        bk_scope2localids[scope] = [
                            torch.zeros([0], dtype = torch.long).to(self.partitioned_nids[0].device) for _ in range(self.num_bk_partitions)
                        ]

                    for partition_id in range(self.num_bk_partitions):
                        u_cids = self.partitioned_u_cids[partition_id]
                        partition_local_ids = torch.where((u_cids >= s_nid) & (u_cids < e_nid))[0]

                        bk_scope2localids[scope][partition_id] = torch.cat(
                            (bk_scope2localids[scope][partition_id], partition_local_ids), dim = 0
                        )

            self.fw_scope2localids = fw_scope2localids
            self.bk_scope2localids = bk_scope2localids

        return prod_scope_eleids



from __future__ import annotations

import torch
import functools

from pyjuice.utils.context_manager import _DecoratorContextManager
from pyjuice.utils.kmeans import faiss_kmeans
from pyjuice.nodes import CircuitNodes, InputNodes, ProdNodes, SumNodes
from .lvd_backend.counting import lvd_by_counting


class LVDistiller(_DecoratorContextManager):
    def __init__(self, discretizer = "KMeans", pseudocount = 0.1, 
                 backend = "counting", verbose = False, **kwargs):

        self.discretizer = discretizer
        self.pseudocount = pseudocount
        self.backend = backend
        self.verbose = verbose
        self.kwargs = kwargs

        self.lv_datasets = []
        self.obs_datasets = []

        self.ns2lv_dataset_id = dict()
        self.ns2obs_dataset_id = dict()

        self.callback = functools.partial(lvd_callback_fn, lvdistiller = self)

    def __enter__(self) -> None:
        CircuitNodes.INIT_CALLBACKS.append(self.callback)

    def __exit__(self, exc_type: Any, exc_value: Any, traceback: Any) -> None:
        CircuitNodes.INIT_CALLBACKS.remove(self.callback)


def lvd_callback_fn(ns: CircuitNodes, lvdistiller: LVDistiller, lv_dataset: Optional[torch.Tensor] = None, 
                    obs_dataset: Optional[torch.Tensor] = None, **kwargs):

    if obs_dataset is not None:
        obs_dataset_id = len(lvdistiller.obs_datasets)
        lvdistiller.obs_datasets.append(obs_dataset)
        lvdistiller.ns2obs_dataset_id[ns] = obs_dataset_id

    if lv_dataset is None:
        # Do nothing if no LV dataset is provided
        return

    # Preprocess LV dataset
    if lv_dataset.dtype == torch.long and lv_dataset.dim() > 1:
        raise ValueError(f"Expecting `lv_dataset` with dtype `torch.long` to be a 1-D tensor, but got a tensor of shape {lv_dataset.size()}.")

    elif lv_dataset.dtype != torch.long:
        if lv_dataset.dim() != 2:
            raise ValueError(f"Expecting `lv_dataset` with dtype `{lv_dataset.dtype}` to be a 2-D tensor, but received a {lv_dataset.dim()}-D tensor.")

        # Discretize `lv_dataset`
        if lvdistiller.discretizer == "KMeans":
            lv_dataset = faiss_kmeans(lv_dataset, n_clusters = ns.num_nodes)
        else:
            raise NotImplementedError(f"Unknown discretizer {lvdistiller.discretizer}.")

    assert lv_dataset.max() < ns.num_nodes, f"Got more latent clusters ({lv_dataset.max().item()}) than nodes ({ns.num_nodes})."

    lv_dataset_id = len(lvdistiller.lv_datasets)
    lvdistiller.lv_datasets.append(lv_dataset)

    lvdistiller.ns2lv_dataset_id[ns] = lv_dataset_id

    # Perform LVD by the specified backend
    if lvdistiller.backend == "counting":
        lvd_by_counting(lvdistiller, ns)
        
    else:
        raise NotImplementedError(f"Unknown LVD backend {lvdistiller.backend}.")


from __future__ import annotations

import torch
import triton
import triton.language as tl

from typing import Optional, Any, Union

from .distributions import Distribution


class Literal(Distribution):
    """
    A class representing Literal (indicator) distributions.
    """
    def __init__(self, lit: Union[bool,int]):
        super(Literal, self).__init__()

        self.lit = int(lit) # Convert True/False to 1/0

    def get_signature(self):
        """
        Get the signature of the current distribution.
        """
        return "Literal"

    def get_metadata(self):
        """
        Get the metadata of the current distribution.
        """
        return [self.lit]

    def num_parameters(self):
        """
        The number of parameters per node.
        """
        return 0

    def num_param_flows(self):
        """
        The number of parameter flows per node.
        """
        return 0

    def init_parameters(self, num_nodes: int, perturbation: float = 2.0, params: Optional[Any] = None, **kwargs):
        """
        Initialize parameters for `num_nodes` nodes.
        Returned parameters should be flattened into a vector.
        """

        return torch.zeros(0)

    @staticmethod
    def fw_mar_fn(local_offsets, data, params_ptr, s_pids, metadata_ptr, s_mids_ptr, mask, num_vars_per_node, BLOCK_SIZE):
        s_mids = tl.load(s_mids_ptr + local_offsets, mask = mask, other = 0)
        lit = tl.load(metadata_ptr + s_mids, mask = mask, other = 0).to(tl.int64)

        probs = tl.where(data == lit, 1.0, 0.0)
        log_probs = tl.log(probs)

        return log_probs

    @staticmethod
    def bk_flow_fn(local_offsets, ns_offsets, data, flows, node_mars_ptr, params_ptr, param_flows_ptr, s_pids, s_pfids, metadata_ptr, 
                   s_mids_ptr, mask, num_vars_per_node, BLOCK_SIZE):
        pass

    @staticmethod
    def sample_fn(samples_ptr, local_offsets, batch_offsets, vids, s_pids, params_ptr, metadata_ptr, s_mids_ptr, mask, batch_size, BLOCK_SIZE, seed):
        pass

    @staticmethod
    def em_fn(local_offsets, params_ptr, param_flows_ptr, s_pids, s_pfids, metadata_ptr, s_mids_ptr, mask,
              step_size, pseudocount, BLOCK_SIZE):
        pass

    def _get_constructor(self):
        return Literal, {"lit": self.lit}


from .distributions import Distribution
from .categorical import Categorical
from .literal import Literal
from .bernoulli import Bernoulli
from .gaussian import Gaussian
from .discrete_logistic import DiscreteLogistic
from .masked_categorical import MaskedCategorical
from .external import External


from __future__ import annotations

from typing import List
from copy import deepcopy

class BitSet(object):
    def __init__(self, byte_length = 1):
        self.values = bytearray(b"\x00" * byte_length)
        self.byte_length = byte_length

        self.length = 0

    @classmethod
    def from_array(cls, arr: List):
        b = BitSet(1)
        for value in arr:
            b.add(value)

        return b

    def to_list(self):
        vals = []
        for i in range(self.byte_length):
            for j in range(8):
                if (self.values[i] & (1 << (7 - j))) != 0:
                    vals.append(i * 8 + j)
        
        return vals

    def add(self, value):
        if self.hasitem(value):
            return

        if value >= self.byte_length * 8:
            new_length = (value // 8 + 1)
            self.values += bytearray(b"\x00" * (new_length - self.byte_length))
            self.byte_length = new_length

        self.values[value // 8] |= (1 << (7 - value % 8))

        self.length += 1

    def remove(self, value):
        if not self.hasitem(value):
            return

        self.values[value // 8] &= (0xff ^ (1 << (7 - value % 8)))

        self.length -= 1

    def hasitem(self, value):
        if value >= self.byte_length * 8:
            return False

        return (self.values[value // 8] & (1 << (7 - value % 8))) != 0

    def contains_all(self, other: BitSet):
        return len(self & other) == len(other)

    def contains_any(self, other: BitSet):
        return len(self & other) > 0

    def __and__(self, other: BitSet):
        if self.byte_length > other.byte_length:
            b = deepcopy(self)
            for i in range(other.byte_length):
                b.values[i] &= other.values[i]
            for i in range(other.byte_length, self.byte_length):
                b.values[i] = 0
        else:
            b = deepcopy(other)
            for i in range(self.byte_length):
                b.values[i] &= self.values[i]
            for i in range(self.byte_length, other.byte_length):
                b.values[i] = 0

        b.length = BitSet._count_ones(b.values)

        return b

    def __or__(self, other: BitSet):
        max_byte_length = max(self.byte_length, other.byte_length)
        b = BitSet(max_byte_length)

        for i in range(self.byte_length):
            b.values[i] |= self.values[i]
        for i in range(other.byte_length):
            b.values[i] |= other.values[i]

        b.length = BitSet._count_ones(b.values)

        return b

    def __eq__(self, other: BitSet):
        min_byte_length = min(self.byte_length, other.byte_length)
        for i in range(min_byte_length):
            if self.values[i] != other.values[i]:
                return False

        if self.byte_length > other.byte_length:
            for i in range(min_byte_length, self.byte_length):
                if self.values[i] != 0:
                    return False
        elif self.byte_length < other.byte_length:
            for i in range(min_byte_length, other.byte_length):
                if other.values[i] != 0:
                    return False

        return True

    def __iter__(self):
        for i in range(self.byte_length):
            for j in range(8):
                if (self.values[i] & (1 << (7 - j))) != 0:
                    yield i * 8 + j

    @staticmethod
    def _count_ones(arr: bytearray):
        count = 0
        for i in range(len(arr)):
            u = arr[i]
            u_count = u - ((u >> 1) & 3681400539) - ((u >> 2) & 1227133513)
            count += ((u_count + (u_count >> 3)) & 3340530119) % 63

        return count

    def __len__(self):
        return self.length

    def __repr__(self):
        if self.length <= 16:
            return "BitSet([" + ",".join([str(v) for v in self]) + "])"
        else:
            return "BitSet(num_elements={})".format(self.length)

    def __hash__(self):
        return hash(bytes(self.values))


from __future__ import annotations

import torch
import triton
import triton.language as tl

from typing import Optional, Any

from .distributions import Distribution


class MaskedCategorical(Distribution):
    """
    A class representing Categorical distributions with masks.

    :param num_cats: number of categories
    :type num_cats: int

    :param mask_mode: type of mask; should be in ["range", "full_mask", "rev_range"]
    :type num_cats: str
    """

    def __init__(self, num_cats: int, mask_mode: str):
        super(MaskedCategorical, self).__init__()

        # `range``: only values in [lo, hi) are accepted
        # `full_mask`: specify a full mask of size Size([num_cats]); 
        #              1 for allowed values and 0 for disallowed
        # `rev_range`: values in [lo, hi) are masked out
        assert mask_mode in ["range", "full_mask", "rev_range"]

        self.num_cats = num_cats
        self.mask_mode = mask_mode

        if self.mask_mode == "range":
            self.fw_mar_fn = self.fw_mar_fn_range
            self.bk_flow_fn = self.bk_flow_fn_range
            self.sample_fn = self.sample_fn_range
            self.em_fn = self.em_fn_range
        elif self.mask_mode == "full_mask":
            self.fw_mar_fn = self.fw_mar_fn_full_mask
            self.bk_flow_fn = self.bk_flow_fn_full_mask
            self.sample_fn = self.sample_fn_full_mask
            self.em_fn = self.em_fn_full_mask
        elif self.mask_mode == "rev_range":
            self.fw_mar_fn = self.fw_mar_fn_rev_range
            self.bk_flow_fn = self.bk_flow_fn_rev_range
            self.sample_fn = self.sample_fn_rev_range
            self.em_fn = self.em_fn_rev_range

    def get_signature(self):
        """
        Get the signature of the current distribution.
        """
        return f"MaskedCategorical-{self.mask_mode}"

    def get_metadata(self):
        """
        Get the metadata of the current distribution.
        """
        return [self.num_cats]

    def num_parameters(self):
        """
        The number of parameters per node.
        """
        if self.mask_mode == "range":
            return self.num_cats + 3
        elif self.mask_mode == "full_mask":
            return self.num_cats * 2 + 1
        elif self.mask_mode == "rev_range":
            return self.num_cats + 3
        else:
            raise ValueError(f"Unknown mask mode {self.mask_mode}.")

    def num_param_flows(self):
        """
        The number of parameter flows per node.
        """
        return self.num_cats

    def init_parameters(self, num_nodes: int, perturbation: float = 2.0, params: Optional[torch.Tensor] = None, **kwargs):
        """
        Initialize parameters for `num_nodes` nodes.
        Returned parameters should be flattened into a vector.
        """
        assert params is not None, "Parameters should be provided to get meta-parameters."
        params = params.reshape(-1, self.num_parameters())
        assert params.size(0) == num_nodes

        num_nodes = params.size(0)

        if self.mask_mode == "range":
            mask_tensor = params[:,self.num_cats:self.num_cats+2]
        elif self.mask_mode == "full_mask":
            mask_tensor = params[:,self.num_cats:self.num_cats*2]
        elif self.mask_mode == "rev_range":
            mask_tensor = params[:,self.num_cats:self.num_cats+2]

        cat_params = torch.exp(torch.rand([num_nodes, self.num_cats]) * -perturbation)
        
        # Apply mask
        self._apply_mask(cat_params, num_nodes, mask_tensor)
        
        cat_params /= cat_params.sum(dim = 1, keepdim = True)

        params = params.clone()
        params[:,:self.num_cats] = cat_params

        return params.reshape(-1)

    def normalize_parameters(self, params: torch.Tensor, mask_tensor: Optional[torch.Tensor] = None):
        params = params.reshape(-1, self.num_parameters())
        num_nodes = params.size(0)

        cat_params = params[:,:self.num_cats]

        if mask_tensor is None:
            if self.mask_mode == "range":
                mask_tensor = params[:,self.num_cats:self.num_cats+2]
            elif self.mask_mode == "full_mask":
                mask_tensor = params[:,self.num_cats:self.num_cats*2]
            elif self.mask_mode == "rev_range":
                mask_tensor = params[:,self.num_cats:self.num_cats+2]

        # Apply mask
        self._apply_mask(cat_params, num_nodes, mask_tensor)

        cat_params /= cat_params.sum(dim = 1, keepdim = True)
        params[:,:self.num_cats] = cat_params

        return params.reshape(-1)

    def set_meta_parameters(self, num_nodes: int, **kwargs):
        assert "mask" in kwargs, "`MaskedCategorical` requires an input argument `mask`."
        mask_tensor = kwargs["mask"]
        assert mask_tensor.size(0) == num_nodes

        if self.mask_mode == "range":
            assert mask_tensor.size(1) == 2
            num_free_cats = mask_tensor[:,1:2] - mask_tensor[:,0:1]
        elif self.mask_mode == "full_mask":
            assert mask_tensor.size(1) == self.num_cats
            num_free_cats = mask_tensor.sum(dim = 1).unsqueeze(1)
        elif self.mask_mode == "rev_range":
            assert mask_tensor.size(1) == 2
            num_free_cats = self.num_cats - (mask_tensor[:,1:2] - mask_tensor[:,0:1])

        cat_params = torch.zeros([num_nodes, self.num_cats])

        params = torch.cat(
            (cat_params, mask_tensor, num_free_cats), 
            dim = 1
        ).contiguous()

        return params.reshape(-1)

    @property
    def need_meta_parameters(self):
        """
        A flag indicating whether users need to pass in meta-parameters to the 
        constructor of InputNodes. In this case, we need to provide information
        regarding the categorical mask.
        """
        return True

    def get_data_dtype(self):
        """
        Get the data dtype for the distribution.
        """
        return torch.long

    def _apply_mask(self, cat_params: torch.Tensor, num_nodes: int, mask_tensor: torch.Tensor):
        if self.mask_mode == "range":
            mask = torch.arange(self.num_cats).unsqueeze(0).expand(num_nodes, -1)
            cat_params[(mask < mask_tensor[:,:1]) | (mask >= mask_tensor[:,1:])] = 0.0
        elif self.mask_mode == "full_mask":
            cat_params[(mask_tensor < 0.5)] = 0.0
        elif self.mask_mode == "rev_range":
            mask = torch.arange(self.num_cats).unsqueeze(0).expand(num_nodes, -1)
            cat_params[(mask >= mask_tensor[:,:1]) & (mask < mask_tensor[:,1:])] = 0.0
        else:
            raise ValueError(f"Unknown mask mode {self.mask_mode}.")

    @staticmethod
    def fw_mar_fn_range(local_offsets, data, params_ptr, s_pids, metadata_ptr, s_mids_ptr, mask, num_vars_per_node, BLOCK_SIZE):
        # Get `num_cats` from `metadata`
        s_mids = tl.load(s_mids_ptr + local_offsets, mask = mask, other = 0)
        num_cats = tl.load(metadata_ptr + s_mids, mask = mask, other = 0).to(tl.int64)

        # Range mode
        lo_idx = s_pids + num_cats
        lo = tl.load(params_ptr + lo_idx, mask = mask, other = 0)
        hi_idx = lo_idx + 1
        hi = tl.load(params_ptr + hi_idx, mask = mask, other = 0)

        param_idx = s_pids + data
        probs = tl.load(params_ptr + param_idx, mask = mask, other = 0)
        log_probs = tl.where((data >= lo) & (data < hi), tl.log(probs), -23.0258509299) # log(1e-10)

        return log_probs

    @staticmethod
    def fw_mar_fn_full_mask(local_offsets, data, params_ptr, s_pids, metadata_ptr, s_mids_ptr, mask, num_vars_per_node, BLOCK_SIZE):
        # Get `num_cats` and from `metadata`
        s_mids = tl.load(s_mids_ptr + local_offsets, mask = mask, other = 0)
        num_cats = tl.load(metadata_ptr + s_mids, mask = mask, other = 0).to(tl.int64)

        # Full mask mode
        prob_mask_idx = s_pids + num_cats + data
        prob_mask = tl.load(params_ptr + prob_mask_idx, mask = mask, other = 0)

        param_idx = s_pids + data
        probs = tl.load(params_ptr + param_idx, mask = mask, other = 0)

        log_probs = tl.where(prob_mask > 0.5, tl.log(probs), -23.0258509299) # log(1e-10)

        return log_probs

    @staticmethod
    def fw_mar_fn_rev_range(local_offsets, data, params_ptr, s_pids, metadata_ptr, s_mids_ptr, mask, num_vars_per_node, BLOCK_SIZE):
        # Get `num_cats` from `metadata`
        s_mids = tl.load(s_mids_ptr + local_offsets, mask = mask, other = 0)
        num_cats = tl.load(metadata_ptr + s_mids, mask = mask, other = 0).to(tl.int64)

        # Rev range mode
        lo_idx = s_pids + num_cats
        lo = tl.load(params_ptr + lo_idx, mask = mask, other = 0)
        hi_idx = lo_idx + 1
        hi = tl.load(params_ptr + hi_idx, mask = mask, other = 0)

        param_idx = s_pids + data
        probs = tl.load(params_ptr + param_idx, mask = mask, other = 0)
        log_probs = tl.where((data < lo) | (data >= hi), tl.log(probs), -23.0258509299) # log(1e-10)

        return log_probs

    @staticmethod
    def bk_flow_fn_range(local_offsets, ns_offsets, data, flows, node_mars_ptr, params_ptr, param_flows_ptr, s_pids, s_pfids, metadata_ptr, 
                         s_mids_ptr, mask, num_vars_per_node, BLOCK_SIZE):
        # Get `num_cats` and from `metadata`
        s_mids = tl.load(s_mids_ptr + local_offsets, mask = mask, other = 0)
        num_cats = tl.load(metadata_ptr + s_mids, mask = mask, other = 0).to(tl.int64)

        # Range mode
        lo_idx = s_pids + num_cats
        lo = tl.load(params_ptr + lo_idx, mask = mask, other = 0)
        hi_idx = lo_idx + 1
        hi = tl.load(params_ptr + hi_idx, mask = mask, other = 0)

        mask = mask & (data >= lo) & (data < hi)

        pf_offsets = s_pfids + data
        tl.atomic_add(param_flows_ptr + pf_offsets, flows, mask = mask)

    @staticmethod
    def bk_flow_fn_full_mask(local_offsets, ns_offsets, data, flows, node_mars_ptr, params_ptr, param_flows_ptr, s_pids, s_pfids, metadata_ptr, 
                             s_mids_ptr, mask, num_vars_per_node, BLOCK_SIZE):
        # Get `num_cats` from `metadata`
        s_mids = tl.load(s_mids_ptr + local_offsets, mask = mask, other = 0)
        num_cats = tl.load(metadata_ptr + s_mids, mask = mask, other = 0).to(tl.int64)

        # Full mask mode
        prob_mask_idx = s_pids + num_cats + data
        prob_mask = tl.load(params_ptr + prob_mask_idx, mask = mask, other = 0)

        mask = mask & (prob_mask > 0.5)

        pf_offsets = s_pfids + data
        tl.atomic_add(param_flows_ptr + pf_offsets, flows, mask = mask)

    @staticmethod
    def bk_flow_fn_rev_range(local_offsets, ns_offsets, data, flows, node_mars_ptr, params_ptr, param_flows_ptr, s_pids, s_pfids, metadata_ptr, 
                             s_mids_ptr, mask, num_vars_per_node, BLOCK_SIZE):
        # Get `num_cats` and from `metadata`
        s_mids = tl.load(s_mids_ptr + local_offsets, mask = mask, other = 0)
        num_cats = tl.load(metadata_ptr + s_mids, mask = mask, other = 0).to(tl.int64)

        # Rev range mode
        lo_idx = s_pids + num_cats
        lo = tl.load(params_ptr + lo_idx, mask = mask, other = 0)
        hi_idx = lo_idx + 1
        hi = tl.load(params_ptr + hi_idx, mask = mask, other = 0)

        mask = mask & ((data < lo) | (data >= hi))

        pf_offsets = s_pfids + data
        tl.atomic_add(param_flows_ptr + pf_offsets, flows, mask = mask)

    @staticmethod
    def sample_fn_range(samples_ptr, local_offsets, batch_offsets, vids, s_pids, params_ptr, metadata_ptr, s_mids_ptr, mask, batch_size, BLOCK_SIZE, seed):
        # Get `num_cats` from `metadata`
        s_mids = tl.load(s_mids_ptr + local_offsets, mask = mask, other = 0)
        num_cats = tl.load(metadata_ptr + s_mids, mask = mask, other = 0).to(tl.int64)

        max_num_cats = tl.max(num_cats, axis = 0)

        rnd_val = tl.rand(seed, tl.arange(0, BLOCK_SIZE))
        sampled_id = tl.zeros([BLOCK_SIZE], dtype = tl.int64) - 1

        # Range mode
        lo_idx = s_pids + num_cats
        lo = tl.load(params_ptr + lo_idx, mask = mask, other = 0)
        hi_idx = lo_idx + 1
        hi = tl.load(params_ptr + hi_idx, mask = mask, other = 0)

        # Sample by computing cumulative probability
        cum_param = tl.zeros([BLOCK_SIZE], dtype = tl.float32)
        for cat_id in range(max_num_cats):
            cat_mask = mask & (cat_id >= lo) & (cat_id < hi)

            param = tl.load(params_ptr + s_pids + cat_id, mask = cat_mask, other = 0)
            cum_param += param

            sampled_id = tl.where((cum_param >= rnd_val) & (sampled_id == -1), cat_id, sampled_id)

        sampled_id = tl.where((sampled_id == -1), 0, sampled_id)

        # Write back to `samples`
        sample_offsets = vids * batch_size + batch_offsets
        tl.store(samples_ptr + sample_offsets, sampled_id, mask = mask)

    @staticmethod
    def sample_fn_full_mask(samples_ptr, local_offsets, batch_offsets, vids, s_pids, params_ptr, metadata_ptr, s_mids_ptr, mask, batch_size, BLOCK_SIZE, seed):
        # Get `num_cats` from `metadata`
        s_mids = tl.load(s_mids_ptr + local_offsets, mask = mask, other = 0)
        num_cats = tl.load(metadata_ptr + s_mids, mask = mask, other = 0).to(tl.int64)

        max_num_cats = tl.max(num_cats, axis = 0)

        rnd_val = tl.rand(seed, tl.arange(0, BLOCK_SIZE))
        sampled_id = tl.zeros([BLOCK_SIZE], dtype = tl.int64) - 1

        # Sample by computing cumulative probability
        cum_param = tl.zeros([BLOCK_SIZE], dtype = tl.float32)
        for cat_id in range(max_num_cats):
            cat_mask = mask & (cat_id < num_cats)

            # Full mask mode
            prob_mask_idx = s_pids + num_cats + cat_id
            prob_mask = tl.load(params_ptr + prob_mask_idx, mask = mask, other = 0)

            param = tl.load(params_ptr + s_pids + cat_id, mask = cat_mask, other = 0)
            cum_param += tl.where(prob_mask > 0.5, param, 0.0)

            sampled_id = tl.where((cum_param >= rnd_val) & (sampled_id == -1), cat_id, sampled_id)

        sampled_id = tl.where((sampled_id == -1), 0, sampled_id)

        # Write back to `samples`
        sample_offsets = vids * batch_size + batch_offsets
        tl.store(samples_ptr + sample_offsets, sampled_id, mask = mask)

    @staticmethod
    def sample_fn_rev_range(samples_ptr, local_offsets, batch_offsets, vids, s_pids, params_ptr, metadata_ptr, s_mids_ptr, mask, batch_size, BLOCK_SIZE, seed):
        # Get `num_cats` from `metadata`
        s_mids = tl.load(s_mids_ptr + local_offsets, mask = mask, other = 0)
        num_cats = tl.load(metadata_ptr + s_mids, mask = mask, other = 0).to(tl.int64)

        max_num_cats = tl.max(num_cats, axis = 0)

        rnd_val = tl.rand(seed, tl.arange(0, BLOCK_SIZE))
        sampled_id = tl.zeros([BLOCK_SIZE], dtype = tl.int64) - 1

        # Range mode
        lo_idx = s_pids + num_cats
        lo = tl.load(params_ptr + lo_idx, mask = mask, other = 0)
        hi_idx = lo_idx + 1
        hi = tl.load(params_ptr + hi_idx, mask = mask, other = 0)

        # Sample by computing cumulative probability
        cum_param = tl.zeros([BLOCK_SIZE], dtype = tl.float32)
        for cat_id in range(max_num_cats):
            cat_mask = mask & ((cat_id < lo) | (cat_id >= hi))

            param = tl.load(params_ptr + s_pids + cat_id, mask = cat_mask, other = 0)
            cum_param += param

            sampled_id = tl.where((cum_param >= rnd_val) & (sampled_id == -1), cat_id, sampled_id)

        sampled_id = tl.where((sampled_id == -1), 0, sampled_id)

        # Write back to `samples`
        sample_offsets = vids * batch_size + batch_offsets
        tl.store(samples_ptr + sample_offsets, sampled_id, mask = mask)

    @staticmethod
    def em_fn_range(local_offsets, params_ptr, param_flows_ptr, s_pids, s_pfids, metadata_ptr, s_mids_ptr, mask,
                    step_size, pseudocount, BLOCK_SIZE):
        # Get `num_cats` from `metadata`
        s_mids = tl.load(s_mids_ptr + local_offsets, mask = mask, other = 0)
        num_cats = tl.load(metadata_ptr + s_mids, mask = mask, other = 0).to(tl.int64)

        eff_num_cats = tl.load(params_ptr + s_pids + num_cats + 2, mask = mask, other = 1)

        max_num_cats = tl.max(num_cats, axis = 0)

        # Compute cumulative flows
        cum_flow = tl.zeros([BLOCK_SIZE], dtype = tl.float32)
        for cat_id in range(max_num_cats):
            cat_mask = mask & (cat_id < num_cats)

            flow = tl.load(param_flows_ptr + s_pfids + cat_id, mask = cat_mask, other = 0)
            cum_flow += flow

        # Parameter update
        numerate_pseudocount = pseudocount / eff_num_cats
        cum_flow += pseudocount
        for cat_id in range(max_num_cats):
            cat_mask = mask & (cat_id < num_cats)

            param = tl.load(params_ptr + s_pids + cat_id, mask = cat_mask, other = 0)
            flow = tl.load(param_flows_ptr + s_pfids + cat_id, mask = cat_mask, other = 0)

            new_param = (1.0 - step_size) * param + step_size * (flow + numerate_pseudocount) / cum_flow
            tl.store(params_ptr + s_pids + cat_id, new_param, mask = cat_mask)

    @staticmethod
    def em_fn_full_mask(local_offsets, params_ptr, param_flows_ptr, s_pids, s_pfids, metadata_ptr, s_mids_ptr, mask,
                        step_size, pseudocount, BLOCK_SIZE):
        # Get `num_cats` from `metadata`
        s_mids = tl.load(s_mids_ptr + local_offsets, mask = mask, other = 0)
        num_cats = tl.load(metadata_ptr + s_mids, mask = mask, other = 0).to(tl.int64)

        eff_num_cats = tl.load(params_ptr + s_pids + num_cats * 2, mask = mask, other = 1)

        max_num_cats = tl.max(num_cats, axis = 0)

        # Compute cumulative flows
        cum_flow = tl.zeros([BLOCK_SIZE], dtype = tl.float32)
        for cat_id in range(max_num_cats):
            cat_mask = mask & (cat_id < num_cats)

            flow = tl.load(param_flows_ptr + s_pfids + cat_id, mask = cat_mask, other = 0)
            cum_flow += flow

        # Parameter update
        numerate_pseudocount = pseudocount / eff_num_cats
        cum_flow += pseudocount
        for cat_id in range(max_num_cats):
            cat_mask = mask & (cat_id < num_cats)

            param = tl.load(params_ptr + s_pids + cat_id, mask = cat_mask, other = 0)
            flow = tl.load(param_flows_ptr + s_pfids + cat_id, mask = cat_mask, other = 0)

            new_param = (1.0 - step_size) * param + step_size * (flow + numerate_pseudocount) / cum_flow
            tl.store(params_ptr + s_pids + cat_id, new_param, mask = cat_mask)

    @staticmethod
    def em_fn_rev_range(local_offsets, params_ptr, param_flows_ptr, s_pids, s_pfids, metadata_ptr, s_mids_ptr, mask,
                        step_size, pseudocount, BLOCK_SIZE):
        # Get `num_cats` from `metadata`
        s_mids = tl.load(s_mids_ptr + local_offsets, mask = mask, other = 0)
        num_cats = tl.load(metadata_ptr + s_mids, mask = mask, other = 0).to(tl.int64)

        eff_num_cats = tl.load(params_ptr + s_pids + num_cats + 2, mask = mask, other = 1)

        max_num_cats = tl.max(num_cats, axis = 0)

        # Compute cumulative flows
        cum_flow = tl.zeros([BLOCK_SIZE], dtype = tl.float32)
        for cat_id in range(max_num_cats):
            cat_mask = mask & (cat_id < num_cats)

            flow = tl.load(param_flows_ptr + s_pfids + cat_id, mask = cat_mask, other = 0)
            cum_flow += flow

        # Parameter update
        numerate_pseudocount = pseudocount / eff_num_cats
        cum_flow += pseudocount
        for cat_id in range(max_num_cats):
            cat_mask = mask & (cat_id < num_cats)

            param = tl.load(params_ptr + s_pids + cat_id, mask = cat_mask, other = 0)
            flow = tl.load(param_flows_ptr + s_pfids + cat_id, mask = cat_mask, other = 0)

            new_param = (1.0 - step_size) * param + step_size * (flow + numerate_pseudocount) / cum_flow
            tl.store(params_ptr + s_pids + cat_id, new_param, mask = cat_mask)

    def _get_constructor(self):
        return MaskedCategorical, {"num_cats": self.num_cats, "mask_mode": self.mask_mode}



from __future__ import annotations

import torch
import random
from copy import deepcopy
from functools import reduce

import pyjuice.transformations as jtf
from typing import Tuple, Sequence, Optional, Type, Dict, Union, Callable
from pyjuice.nodes import multiply, summate, inputs, set_block_size
from pyjuice.nodes.distributions import *
from pyjuice.structures.hclt import HCLT
from pyjuice.utils import BitSet
from pyjuice.utils.util import max_cdf_power_of_2


def PD(data_shape: Tuple, num_latents: int, 
       split_intervals: Optional[Union[int, Tuple[int]]] = None, 
       split_points: Optional[Sequence[Sequence[int]]] = None,
       max_split_depth: Optional[int] = None,
       max_prod_block_conns: int = 4,
       structure_type: str = "sum_dominated",
       input_ns_fn: Optional[Callable] = None,
       input_dist: Optional[Distribution] = None,
       input_node_type: Type[Distribution] = Categorical, 
       input_node_params: Dict = {"num_cats": 256},
       tie_homogeneous_params: bool = False,
       block_size: Optional[int] = None):
    """
    Generate PCs with the PD structure (https://arxiv.org/pdf/1202.3732.pdf).

    :param data_shape: shape of the data (e.g., [H, W, 3] for images and [S] for sequences)
    :type data_shape: Tuple

    :param num_latents: size of the latent space
    :type num_latents: int

    :param split_intervals: a sequence of integers specifying the interval between split points in every dimension; either this or `split_points` needs to be specified
    :type split_intervals: Optional[Union[int, Tuple[int]]]

    :param split_points: a sequence of split points in each dimension; either this or `split_intervals` needs to be specified
    :type split_points: Optional[Sequence[Sequence[int]]]

    :param max_split_depth: maximum depth of the splits
    :type max_split_depth: Optional[int]

    :param max_prod_block_conns: the maximum number of product nodes connected to every sum node
    :type max_prod_block_conns: int

    :param structure_type: whether to reuse sum nodes; no reuse: "sum_dominated", reuse: "prod_dominated"
    :type structure_type: str

    :param input_dist: input distribution
    :type input_dist: Distribution

    :param tie_homogeneous_params: whether to tie parameters of sum/input nodes with compatible structures
    :type tie_homogeneous_params: bool

    :param block_size: block size
    :type block_size: int
    """
    assert structure_type in ["sum_dominated", "prod_dominated"]

    if input_dist is not None:
        input_node_type, input_node_params = input_dist._get_constructor()

    # Specify block size
    if block_size is None:
        if num_latents <= 32:
            block_size = min(16, max_cdf_power_of_2(num_latents))
        else:
            block_size = min(32, max_cdf_power_of_2(num_latents))

    num_node_blocks = num_latents // block_size

    num_axes = len(data_shape)

    # A dictionary of source nodes
    source_ns_dict = dict()

    # Construct split points
    if split_intervals is not None:
        if isinstance(split_intervals, int):
            split_intervals = (split_intervals,) * num_axes

        split_points = [
            [j for j in range(split_intervals[axis], data_shape[axis], split_intervals[axis])] 
            for axis in range(num_axes)
        ]
    else:
        assert split_points is not None
        assert len(split_points) == num_axes
        for axis in range(num_axes):
            assert all(map(lambda x: 0 < x < data_shape[axis] - 1, split_points[axis]))

    global_split_points = split_points

    # Cache
    hypercube2ns = dict()

    def hypercube2scope(hypercube):
        vars = None
        cum_id = 1
        for axis in range(num_axes - 1, -1, -1):
            if vars is None:
                vars = torch.arange(hypercube[0][axis], hypercube[1][axis])
            else:
                s, e = hypercube[0][axis], hypercube[1][axis]
                vars = vars[None,:].repeat(e - s, 1) + (torch.arange(s, e) * cum_id)[:,None]
                vars = vars.reshape(-1)

            cum_id *= data_shape[axis]

        return BitSet.from_array(vars.tolist())

    def updated_hypercube(hypercube, axis, s = None, e = None):
        hypercube = [list(hypercube[0]), list(hypercube[1])]
        if s is not None:
            hypercube[0][axis] = s
        if e is not None:
            hypercube[1][axis] = e

        hypercube = (tuple(hypercube[0]), tuple(hypercube[1]))
        return hypercube

    def hypercube2shape(hypercube):
        return tuple(hypercube[1][i] - hypercube[0][i] for i in range(len(hypercube[0])))

    def get_signature(hypercube, *ch_ns):
        return (hypercube2shape(hypercube), tuple(len(ns.scope) for ns in ch_ns))

    def create_input_ns(hypercube):
        scope = hypercube2scope(hypercube)
        if input_ns_fn is not None:
            return input_ns_fn(scope, num_latents, block_size)
        else:
            input_nodes = []
            for var in scope:
                if not tie_homogeneous_params:
                    ns = inputs(var, num_node_blocks = num_node_blocks, dist = input_node_type(**input_node_params))
                else:
                    if "input" in source_ns_dict:
                        ns = source_ns_dict["input"].duplicate(var, tie_params = True)
                    else:
                        ns = inputs(var, num_node_blocks = num_node_blocks, dist = input_node_type(**input_node_params))
                        source_ns_dict["input"] = ns
                input_nodes.append(ns)

            edge_ids = torch.arange(0, num_node_blocks)[None,:].repeat(2, 1)
            pns = multiply(*input_nodes)
            if not tie_homogeneous_params:
                return summate(pns, num_node_blocks = num_node_blocks, edge_ids = edge_ids)
            else:
                signature = get_signature(hypercube, pns)
                if signature in source_ns_dict:
                    return source_ns_dict[signature].duplicate(pns, tie_params = True)
                else:
                    ns = summate(pns, num_node_blocks = num_node_blocks, edge_ids = edge_ids)
                    source_ns_dict[signature] = ns
                    return ns

    def recursive_construct(hypercube, depth = 1):
        if hypercube in hypercube2ns:
            return hypercube2ns[hypercube]

        if max_split_depth is not None and depth > max_split_depth:
            ns = create_input_ns(hypercube)
            hypercube2ns[hypercube] = ns
            return ns

        # Try to split over every axis
        pns = []
        for axis in range(num_axes):
            s, e = hypercube[0][axis], hypercube[1][axis]
            split_points = [sp for sp in global_split_points[axis] if s < sp < e]

            for sp in split_points:
                ns1 = recursive_construct(updated_hypercube(hypercube, axis, e = sp))
                ns2 = recursive_construct(updated_hypercube(hypercube, axis, s = sp))

                if structure_type == "sum_dominated":
                    pn = multiply(ns1, ns2)
                elif structure_type == "prod_dominated":
                    pn = multiply(ns1.duplicate(), ns2.duplicate())

                pns.append(pn)

        if len(pns) == 0:
            # No split point found. Create input nodes instead
            ns = create_input_ns(hypercube)
        elif hypercube == root_hypercube:
            ns = summate(*pns, num_node_blocks = 1, block_size = 1)
        else:
            if len(pns) <= max_prod_block_conns:
                if not tie_homogeneous_params:
                    ns = summate(*pns, num_node_blocks = num_node_blocks)
                else:
                    signature = get_signature(hypercube, *pns)
                    if signature in source_ns_dict:
                        ns = source_ns_dict[signature].duplicate(*pns, tie_params = True)
                    else:
                        ns = summate(*pns, num_node_blocks = num_node_blocks)
                        source_ns_dict[signature] = ns
            else:
                block_ids = torch.topk(torch.rand([num_node_blocks, len(pns)]), k = max_prod_block_conns, dim = 1).indices
                par_ids = torch.arange(0, num_node_blocks)[:,None,None].repeat(1, max_prod_block_conns, num_node_blocks)
                chs_ids = block_ids[:,:,None] * num_node_blocks + torch.arange(0, num_node_blocks)[None,None,:]
                edge_ids = torch.stack((par_ids.reshape(-1), chs_ids.reshape(-1)), dim = 0)

                if not tie_homogeneous_params:
                    ns = summate(*pns, num_node_blocks = num_node_blocks, edge_ids = edge_ids)
                else:
                    signature = get_signature(hypercube, *pns)
                    if signature in source_ns_dict:
                        ns = source_ns_dict[signature].duplicate(*pns, tie_params = True)
                    else:
                        ns = summate(*pns, num_node_blocks = num_node_blocks, edge_ids = edge_ids)
                        source_ns_dict[signature] = ns

        hypercube2ns[hypercube] = ns

        return ns

    with set_block_size(block_size = block_size):
        root_hypercube = ((0,) * num_axes, deepcopy(data_shape))
        root_ns = recursive_construct(root_hypercube)

    return root_ns


def PDHCLT(data: torch.Tensor, data_shape: Tuple, num_latents: int, 
           split_intervals: Optional[Union[int, Tuple[int]]] = None, 
           split_points: Optional[Sequence[Sequence[int]]] = None,
           max_split_depth: Optional[int] = None,
           max_prod_block_conns: int = 4,
           structure_type: str = "sum_dominated",
           input_dist: Optional[Distribution] = None,
           input_node_type: Type[Distribution] = Categorical, 
           input_node_params: Dict = {"num_cats": 256},
           hclt_kwargs: Dict = {"num_bins": 32, "sigma": 0.5 / 32, "chunk_size": 32},
           block_size: Optional[int] = None):

    assert data.dim() == 2
    assert data.size(1) == reduce(lambda x, y: x * y, data_shape)

    if input_dist is not None:
        input_node_type, input_node_params = input_dist._get_constructor()

    def input_ns_fn(scope, num_latents, block_size):
        vars = torch.tensor(scope.to_list()).sort().values
        ns = HCLT(
            x = data[:,vars], 
            num_latents = num_latents, 
            input_node_type = input_node_type,
            input_node_params = input_node_params,
            num_root_ns = num_latents,
            block_size = block_size,
            **hclt_kwargs
        )

        # Map input variables
        var_mapping = {v: vars[v].item() for v in range(vars.size(0))}
        ns = jtf.deepcopy(ns, var_mapping = var_mapping)

        return ns

    ns = PD(data_shape = data_shape, num_latents = num_latents, 
            split_intervals = split_intervals, split_points = split_points,
            max_split_depth = max_split_depth, max_prod_block_conns = max_prod_block_conns,
            structure_type = structure_type, input_ns_fn = input_ns_fn,
            input_node_type = input_node_type, input_node_params = input_node_params,
            block_size = block_size)

    if ns.num_node_blocks > 1:
        ns = summate(*ns.chs, num_node_blocks = 1, block_size = 1)

    return ns



from __future__ import annotations

import torch
import pickle
from functools import partial
from typing import Sequence

from pyjuice.model import TensorCircuit
from pyjuice.nodes import CircuitNodes, InputNodes, ProdNodes, SumNodes
from .serialization import serialize_nodes, deserialize_nodes


def save(fname: str, model: Union[CircuitNodes,TensorCircuit]):
    """
    Save a PC to file. The file should end with ".jpc".

    :param fname: file name
    :type fname: str

    :param model: an uncompiled (`CircuitNodes`) or compiled (`TensorCircuit`) PC
    :type model: Union[CircuitNodes,TensorCircuit]
    """
    if isinstance(model, TensorCircuit):
        model.update_parameters()
        root_ns = model.root_ns
    else:
        root_ns = model

    if fname.endswith(".jpc"):
        sel_nodes = serialize_nodes(root_ns)
        with open(fname, "wb") as f:
            pickle.dump(sel_nodes, f)
    else:
        raise ValueError(f"Unknown file type `.{fname.split('.')[-1]}`.")


def load(fname: str) -> CircuitNodes:
    """
    Load a PC from file. The file should end with ".jpc".

    :param fname: file name
    :type fname: str

    :returns: a PC node
    """
    if fname.endswith(".jpc"):
        with open(fname, "rb") as f:
            sel_nodes = pickle.load(f)
        root_ns = deserialize_nodes(sel_nodes)
    else:
        raise ValueError(f"Unknown file type `.{fname.split('.')[-1]}`.")

    return root_ns


from __future__ import annotations

from functools import reduce
from typing import Sequence, Union

from pyjuice.nodes import CircuitNodes
from pyjuice.utils import BitSet


def get_subsumed_scopes(root_ns, scopes: Union[Sequence[BitSet],Sequence[int],BitSet], type: str = "any"):
    """
    Get all scopes in the PC `root_ns` whose scope subsumes `scopes`. `type` could be "any" or "all".
    """
    if not isinstance(root_ns, CircuitNodes):
        root_ns = root_ns.root_nodes

    if isinstance(scopes, BitSet):
        scopes = [scopes]

    if isinstance(scopes[0], int):
        scopes = [BitSet.from_array([var]) for var in scopes]

    target_scopes = set()

    if type == "any":
        for ns in root_ns:
            if ns.scope in target_scopes:
                continue

            if any([ns.scope.contains_all(ref_scope) for ref_scope in scopes]):
                target_scopes.add(ns.scope)

    elif type == "all":
        ref_scope = reduce(lambda x, y: x | y, scopes)
        for ns in root_ns:
            if ns.scope in target_scopes:
                continue

            if ns.scope.contains_any(ref_scope):
                target_scopes.add(ns.scope)

    else:
        raise ValueError(f"Illegal type {type}.")

    return list(target_scopes)



import torch
from torch.autograd import Function


class ReverseGrad(Function):
    @staticmethod
    def forward(ctx, x):
        return x

    @staticmethod
    def backward(ctx, grad_out):
        return -grad_out


class PseudoHookFunc(Function):
    @staticmethod
    def forward(ctx, *args):
        ctx.num_vars = len(args)
        ctx.save_for_backward(*args)
        return args[-1]

    @staticmethod
    def backward(ctx, grad):
        return tuple(grad if i == ctx.num_vars-1 else torch.zeros_like(ctx.saved_tensors[i]) for i in range(ctx.num_vars))



from __future__ import annotations

import torch
import torch.nn as nn
import triton
import triton.language as tl
from numba import njit

from pyjuice.utils.kernel_launcher import triton_jit


@triton_jit
def cum_par_kernel(cum_pflows, mparams, par_start_ids, blk_sizes, blk_intervals, 
                   global_nids, constexprs, BLOCK_ID: tl.constexpr, BLOCK_SIZE: tl.constexpr):

    pid = tl.program_id(axis = 0)

    # Retrieve the constants
    num_blocks = tl.load(constexprs + 1).to(tl.int64)

    offs_m = pid * BLOCK_ID + tl.arange(0, BLOCK_ID)
    mask_m = offs_m < num_blocks

    offs_blk = tl.arange(0, BLOCK_SIZE)

    par_start = tl.load(par_start_ids + offs_m, mask = mask_m, other = 0)
    blk_size = tl.load(blk_sizes + offs_m, mask = mask_m, other = 0)
    blk_interval = tl.load(blk_intervals + offs_m, mask = mask_m, other = 0)
    global_nid = tl.load(global_nids + offs_m, mask = mask_m, other = 0)

    offs_par = par_start[:,None] + offs_blk[None,:] * blk_interval[:,None]
    mask_par = mask_m[:,None] & (offs_blk[None,:] < blk_size[:,None])
    pars = tl.load(mparams + offs_par, mask = mask_par, other = 0)
    sum_pars = tl.sum(pars, axis = 1)

    tl.atomic_add(cum_pflows + global_nid, sum_pars, mask = mask_m)


@triton_jit
def par_update_kernel(mparams, cum_pflows, nchs, par_start_ids, blk_sizes, blk_intervals,
                      global_nids, constexprs, BLOCK_ID: tl.constexpr, BLOCK_SIZE: tl.constexpr):

    pid = tl.program_id(axis = 0)

    # Retrieve the constants
    pseudocount = tl.load(constexprs)
    num_blocks = tl.load(constexprs + 1).to(tl.int64)

    offs_m = pid * BLOCK_ID + tl.arange(0, BLOCK_ID)
    mask_m = offs_m < num_blocks

    offs_blk = tl.arange(0, BLOCK_SIZE)

    par_start = tl.load(par_start_ids + offs_m, mask = mask_m, other = 0)
    blk_size = tl.load(blk_sizes + offs_m, mask = mask_m, other = 0)
    blk_interval = tl.load(blk_intervals + offs_m, mask = mask_m, other = 0)
    global_nid = tl.load(global_nids + offs_m, mask = mask_m, other = 0)

    offs_par = par_start[:,None] + offs_blk[None,:] * blk_interval[:,None]
    mask_par = mask_m[:,None] & (offs_blk[None,:] < blk_size[:,None])
    pars = tl.load(mparams + offs_par, mask = mask_par, other = 0)

    sum_pars = tl.load(cum_pflows + global_nid, mask = mask_m, other = 1)
    nch = tl.load(nchs + global_nid, mask = mask_m, other = 1)

    norm_param = (pars + pseudocount / nch[:,None]) / (sum_pars[:,None] + pseudocount)

    tl.store(mparams + offs_par, norm_param, mask = mask_par)


@njit
def cum_par_numba_kernel(cum_pflows, params, par_start_ids, blk_sizes, blk_intervals, global_nids):
    for i in range(par_start_ids.shape[0]):
        par_start_id = par_start_ids[i]
        blk_size = blk_sizes[i]
        blk_interval = blk_intervals[i]
        global_nid = global_nids[i]

        cum_par = 0.0
        for j in range(blk_size):
            cum_par += params[par_start_id+j*blk_interval]

        cum_pflows[global_nid] += cum_par


@njit
def par_update_numba_kernel(params, cum_pflows, nchs, par_start_ids, blk_sizes, blk_intervals, global_nids, pseudocount):
    for i in range(par_start_ids.shape[0]):
        par_start = par_start_ids[i]
        blk_size = blk_sizes[i]
        blk_interval = blk_intervals[i]
        global_nid = global_nids[i]

        cum_par = cum_pflows[global_nid]
        nch = nchs[global_nid]

        for j in range(blk_size):
            par = params[par_start+j*blk_interval]
            norm_par = (par + pseudocount / nch) / (cum_par + pseudocount)
            params[par_start+j*blk_interval] = norm_par


def normalize_parameters(params, par_update_kwargs, pseudocount: float = 0.0):

    par_start_ids, _, blk_sizes, blk_intervals, global_nids, nchs, cum_pflows, metadata = par_update_kwargs

    tot_num_nodes = metadata["tot_num_nodes"]
    BLOCK_SIZE = metadata["BLOCK_SIZE"]

    if cum_pflows is None:
        cum_pflows = torch.zeros([tot_num_nodes], dtype = torch.float32, device = params.device)
    else:
        cum_pflows[:] = 0.0

    use_cuda = params.is_cuda

    if use_cuda:

        num_blocks = par_start_ids.size(0)
        BLOCK_ID = 2048 // BLOCK_SIZE

        constexprs = torch.tensor([pseudocount, num_blocks]).to(params.device)

        grid = (triton.cdiv(num_blocks, BLOCK_ID),)

        cum_par_kernel[grid](
            cum_pflows, params, par_start_ids, blk_sizes, blk_intervals, 
            global_nids, constexprs, BLOCK_ID, BLOCK_SIZE
        )

        par_update_kernel[grid](
            params, cum_pflows, nchs, par_start_ids, blk_sizes, blk_intervals,
            global_nids, constexprs, BLOCK_ID, BLOCK_SIZE
        )

    else:

        cum_pflows = cum_pflows.numpy()
        params = params.numpy()
        par_start_ids = par_start_ids.numpy()
        blk_sizes = blk_sizes.numpy()
        blk_intervals = blk_intervals.numpy()
        global_nids = global_nids.numpy()
        nchs = nchs.numpy()

        cum_par_numba_kernel(
            cum_pflows, params, par_start_ids, blk_sizes, 
            blk_intervals, global_nids
        )

        par_update_numba_kernel(
            params, cum_pflows, nchs, par_start_ids, blk_sizes, 
            blk_intervals, global_nids, pseudocount
        )



from __future__ import annotations

import numpy as np
import torch
import torch.nn as nn
import triton
import triton.language as tl
from numba import njit
from typing import Sequence

from pyjuice.nodes import CircuitNodes
from pyjuice.utils.kernel_launcher import FastJITFunction


@njit
def _record_par_blks(par_start_ids, pflow_start_ids, blk_sizes, blk_intervals, global_nids, nchs,
                     num_edges_per_ng, ns_num_node_blocks, ns_block_size, cs_block_size, pid, 
                     global_nid, par_start, pflow_start, BLOCK_SIZE):
    for local_ngid in range(ns_num_node_blocks):
        num_edges = num_edges_per_ng[local_ngid]
        num_chs = num_edges * cs_block_size

        for sid in range(0, num_chs, BLOCK_SIZE):
            eid = min(sid + BLOCK_SIZE, num_chs)
            blk_size = eid - sid

            for gid in range(ns_block_size):
                psid = par_start + sid * ns_block_size + gid
                pfsid = pflow_start + sid * ns_block_size + gid
                global_ind = global_nid + gid

                par_start_ids[pid] = par_start + sid * ns_block_size + gid
                pflow_start_ids[pid] = pflow_start + sid * ns_block_size + gid
                blk_sizes[pid] = blk_size
                blk_intervals[pid] = ns_block_size
                global_nids[pid] = global_nid + gid
                nchs[pid] = num_edges * cs_block_size

                pid += 1

        pflow_start += ns_block_size * num_edges * cs_block_size
        par_start += ns_block_size * num_edges * cs_block_size
        global_nid += ns_block_size

    return global_nid, pid


@torch.no_grad()
def compile_par_update_fn(root_ns: CircuitNodes, BLOCK_SIZE: int = 32, buffer_inc_interval: int = 10000, use_numba: bool = True):

    assert BLOCK_SIZE & (BLOCK_SIZE - 1) == 0, "`BLOCK_SIZE` must be a power of 2."

    par_start_ids = np.zeros([buffer_inc_interval], dtype = np.int64)
    pflow_start_ids = np.zeros([buffer_inc_interval], dtype = np.int64)
    blk_sizes = np.zeros([buffer_inc_interval], dtype = np.int64)
    blk_intervals = np.zeros([buffer_inc_interval], dtype = np.int64)
    global_nids = np.zeros([buffer_inc_interval], dtype = np.int64)
    nchs = np.zeros([buffer_inc_interval], dtype = np.int64)
    pid = 0

    global_nid = 0
    for i, ns in enumerate(root_ns):
        if not ns.is_sum() or ns.is_tied():
            continue

        par_start = ns._param_range[0]
        pflow_start = ns._param_flow_range[0]
        tot_n_pars = ns._param_range[1] - ns._param_range[0]

        num_edges_per_ng = torch.bincount(ns.edge_ids[0,:], minlength = ns.num_node_blocks).contiguous().numpy()

        # Enlarge the buffer if needed
        est_num_slots = triton.cdiv(ns.edge_ids.size(1) * ns.block_size * ns.ch_block_size, BLOCK_SIZE) + ns.num_nodes
        if pid + est_num_slots > par_start_ids.shape[0]:
            curr_size = par_start_ids.shape[0]
            inc_shape = triton.cdiv(pid + est_num_slots - curr_size, buffer_inc_interval) * buffer_inc_interval

            par_start_ids_new = np.zeros([curr_size + inc_shape], dtype = np.int64)
            par_start_ids_new[:curr_size] = par_start_ids[:curr_size]
            par_start_ids = par_start_ids_new

            pflow_start_ids_new = np.zeros([curr_size + inc_shape], dtype = np.int64)
            pflow_start_ids_new[:curr_size] = pflow_start_ids[:curr_size]
            pflow_start_ids = pflow_start_ids_new

            blk_sizes_new = np.zeros([curr_size + inc_shape], dtype = np.int64)
            blk_sizes_new[:curr_size] = blk_sizes[:curr_size]
            blk_sizes = blk_sizes_new

            blk_intervals_new = np.zeros([curr_size + inc_shape], dtype = np.int64)
            blk_intervals_new[:curr_size] = blk_intervals[:curr_size]
            blk_intervals = blk_intervals_new

            global_nids_new = np.zeros([curr_size + inc_shape], dtype = np.int64)
            global_nids_new[:curr_size] = global_nids[:curr_size]
            global_nids = global_nids_new

            nchs_new = np.zeros([curr_size + inc_shape], dtype = np.int64)
            nchs_new[:curr_size] = nchs[:curr_size]
            nchs = nchs_new

            buffer_inc_interval *= 2

        if use_numba:
            ns_num_node_blocks = ns.num_node_blocks
            ns_block_size = ns.block_size
            cs_block_size = ns.ch_block_size

            global_nid, pid = _record_par_blks(
                par_start_ids, pflow_start_ids, blk_sizes, blk_intervals, global_nids, nchs,
                num_edges_per_ng, ns_num_node_blocks, ns_block_size, cs_block_size, pid, 
                global_nid, par_start, pflow_start, BLOCK_SIZE
            )

        else:
            ns_gid_range = torch.arange(0, ns.block_size)

            for local_ngid in range(ns.num_node_blocks):
                num_edges = num_edges_per_ng[local_ngid]
                num_chs = num_edges * ns.ch_block_size

                for sid in range(0, num_chs, BLOCK_SIZE):
                    eid = min(sid + BLOCK_SIZE, num_chs)
                    blk_size = eid - sid

                    curr_psids = par_start + sid * ns.block_size + ns_gid_range
                    curr_pfsids = pflow_start + sid * ns.block_size + ns_gid_range
                    curr_global_nids = global_nid + ns_gid_range

                    par_start_ids[pid:pid+ns.block_size] = curr_psids
                    pflow_start_ids[pid:pid+ns.block_size] = curr_pfsids
                    blk_sizes[pid:pid+ns.block_size] = blk_size
                    blk_intervals[pid:pid+ns.block_size] = ns.block_size
                    global_nids[pid:pid+ns.block_size] = curr_global_nids
                    nchs[pid:pid+ns.block_size] = num_edges * ns.ch_block_size

                    pid += ns.block_size

                par_start += ns.block_size * num_edges * ns.ch_block_size
                pflow_start += ns.block_size * num_edges * ns.ch_block_size
                global_nid += ns.block_size

    par_start_ids = torch.from_numpy(par_start_ids[:pid]).contiguous()
    pflow_start_ids = torch.from_numpy(pflow_start_ids[:pid]).contiguous()
    blk_sizes = torch.from_numpy(blk_sizes[:pid]).contiguous()
    blk_intervals = torch.from_numpy(blk_intervals[:pid]).contiguous()
    global_nids = torch.from_numpy(global_nids[:pid]).contiguous()
    nchs = torch.from_numpy(nchs[:pid]).contiguous()

    cum_pflows = torch.zeros([global_nids[-1] + 1], dtype = torch.float32)
    
    metadata = {"tot_num_nodes": global_nids[-1].item() + 1, "BLOCK_SIZE": BLOCK_SIZE}

    return [par_start_ids, pflow_start_ids, blk_sizes, blk_intervals, global_nids, nchs, cum_pflows, metadata]


def par_update_to_device(par_update_kwargs, device):

    par_start_ids, pflow_start_ids, blk_sizes, blk_intervals, global_nids, nchs, cum_pflows, metadata = par_update_kwargs

    return [
        par_start_ids.to(device),
        pflow_start_ids.to(device),
        blk_sizes.to(device),
        blk_intervals.to(device),
        global_nids.to(device),
        nchs.to(device),
        cum_pflows.to(device),
        metadata
    ]


@triton.jit
def cum_pflow_kernel(cum_pflows, mparams, param_flows, nchs, par_start_ids, pflow_start_ids, blk_sizes, blk_intervals, 
                     global_nids, constexprs, keep_zero_params: tl.constexpr, BLOCK_ID: tl.constexpr, 
                     BLOCK_SIZE: tl.constexpr):

    pid = tl.program_id(axis = 0)

    # Retrieve the constants
    pseudocount = tl.load(constexprs + 1)
    num_blocks = tl.load(constexprs + 2).to(tl.int64)

    offs_m = pid * BLOCK_ID + tl.arange(0, BLOCK_ID)
    mask_m = offs_m < num_blocks

    offs_blk = tl.arange(0, BLOCK_SIZE)

    pflow_start = tl.load(pflow_start_ids + offs_m, mask = mask_m, other = 0)
    blk_size = tl.load(blk_sizes + offs_m, mask = mask_m, other = 0)
    blk_interval = tl.load(blk_intervals + offs_m, mask = mask_m, other = 0)
    global_nid = tl.load(global_nids + offs_m, mask = mask_m, other = 0)

    offs_pflow = pflow_start[:,None] + offs_blk[None,:] * blk_interval[:,None]
    mask_pflow = mask_m[:,None] & (offs_blk[None,:] < blk_size[:,None])
    pflows = tl.load(param_flows + offs_pflow, mask = mask_pflow, other = 0)

    if keep_zero_params == 1:
        par_start = tl.load(par_start_ids + offs_m, mask = mask_m, other = 0)
        offs_par = par_start[:,None] + offs_blk[None,:] * blk_interval[:,None]
        old_params = tl.load(mparams + offs_par, mask = mask_pflow, other = 0)

        nch = tl.load(nchs + offs_m, mask = mask_m, other = 1)
        pflows += (pseudocount / nch[:,None])

        nflows = tl.sum(tl.where(old_params < 1e-12, 0.0, pflows), axis = 1)
    else:
        nflows = tl.sum(pflows, axis = 1)

    tl.atomic_add(cum_pflows + global_nid, nflows, mask = mask_m)


@triton.jit
def em_par_update_kernel(mparams, param_flows, cum_pflows, nchs, par_start_ids, pflow_start_ids, blk_sizes, blk_intervals,
                         global_nids, constexprs, keep_zero_params: tl.constexpr, BLOCK_ID: tl.constexpr, 
                         BLOCK_SIZE: tl.constexpr):

    pid = tl.program_id(axis = 0)

    # Retrieve the constants
    step_size = tl.load(constexprs)
    pseudocount = tl.load(constexprs + 1)
    num_blocks = tl.load(constexprs + 2).to(tl.int64)

    offs_m = pid * BLOCK_ID + tl.arange(0, BLOCK_ID)
    mask_m = offs_m < num_blocks

    offs_blk = tl.arange(0, BLOCK_SIZE)

    par_start = tl.load(par_start_ids + offs_m, mask = mask_m, other = 0)
    pflow_start = tl.load(pflow_start_ids + offs_m, mask = mask_m, other = 0)
    blk_size = tl.load(blk_sizes + offs_m, mask = mask_m, other = 0)
    blk_interval = tl.load(blk_intervals + offs_m, mask = mask_m, other = 0)
    global_nid = tl.load(global_nids + offs_m, mask = mask_m, other = 0)

    offs_pflow = pflow_start[:,None] + offs_blk[None,:] * blk_interval[:,None]
    mask_pflow = mask_m[:,None] & (offs_blk[None,:] < blk_size[:,None])
    pflows = tl.load(param_flows + offs_pflow, mask = mask_pflow, other = 0)

    nflows = tl.load(cum_pflows + global_nid, mask = mask_m, other = 1)
    nch = tl.load(nchs + offs_m, mask = mask_m, other = 1)

    if keep_zero_params == 1:
        new_param = (pflows + pseudocount / nch[:,None]) / nflows[:,None]
    else:
        new_param = (pflows + pseudocount / nch[:,None]) / (nflows[:,None] + pseudocount)

    offs_par = par_start[:,None] + offs_blk[None,:] * blk_interval[:,None]
    old_param = tl.load(mparams + offs_par, mask = mask_pflow, other = 0)

    updated_param = (1.0 - step_size) * old_param + step_size * new_param

    if keep_zero_params == 1:
        updated_params = tl.where(old_param < 1e-12, 0.0, updated_param)

    tl.store(mparams + offs_par, updated_param, mask = mask_pflow)


@triton.jit
def sgd_par_update_kernel(mparams, param_grads, par_start_ids, pgrad_start_ids, blk_sizes, blk_intervals,
                         global_nids, constexprs, keep_zero_params: tl.constexpr, BLOCK_ID: tl.constexpr, 
                         BLOCK_SIZE: tl.constexpr):

    pid = tl.program_id(axis = 0)

    # Retrieve the constants
    lr = tl.load(constexprs)
    num_blocks = tl.load(constexprs + 1).to(tl.int64)

    offs_m = pid * BLOCK_ID + tl.arange(0, BLOCK_ID)
    mask_m = offs_m < num_blocks

    offs_blk = tl.arange(0, BLOCK_SIZE)

    par_start = tl.load(par_start_ids + offs_m, mask = mask_m, other = 0)
    pgrad_start = tl.load(pgrad_start_ids + offs_m, mask = mask_m, other = 0)
    blk_size = tl.load(blk_sizes + offs_m, mask = mask_m, other = 0)
    blk_interval = tl.load(blk_intervals + offs_m, mask = mask_m, other = 0)
    global_nid = tl.load(global_nids + offs_m, mask = mask_m, other = 0)

    offs_pgrad = pgrad_start[:,None] + offs_blk[None,:] * blk_interval[:,None]
    mask_pgrad = mask_m[:,None] & (offs_blk[None,:] < blk_size[:,None])
    pgrads = tl.load(param_grads + offs_pgrad, mask = mask_pgrad, other = 0)

    offs_par = par_start[:,None] + offs_blk[None,:] * blk_interval[:,None]
    old_param = tl.load(mparams + offs_par, mask = mask_pgrad, other = 0)

    if keep_zero_params:
        updated_params = tl.where(old_param < 1e-12, 0.0, tl.exp(tl.log(old_param) + lr * pgrads))
    else:
        updated_param = tl.exp(tl.log(old_param) + lr * pgrads)

    tl.store(mparams + offs_par, updated_param, mask = mask_pgrad)


def em_par_update(params: torch.Tensor, param_flows: torch.Tensor, par_update_kwargs: Sequence, 
                  step_size: float, pseudocount: float = 0.0, keep_zero_params: bool = True):

    par_start_ids, pflow_start_ids, blk_sizes, blk_intervals, global_nids, nchs, cum_pflows, metadata = par_update_kwargs

    tot_num_nodes = metadata["tot_num_nodes"]
    BLOCK_SIZE = metadata["BLOCK_SIZE"]

    if cum_pflows is None:
        cum_pflows = torch.zeros([tot_num_nodes], dtype = torch.float32, device = params.device)
    else:
        cum_pflows[:] = 0.0

    num_blocks = par_start_ids.size(0)
    BLOCK_ID = 2048 // BLOCK_SIZE

    grid = (triton.cdiv(num_blocks, BLOCK_ID),)

    constexprs = torch.tensor([step_size, pseudocount, num_blocks]).to(params.device)

    keep_zero_params = 1 if keep_zero_params else 0

    cum_pflow_kernel[grid](
        cum_pflows, params, param_flows, nchs, par_start_ids, pflow_start_ids, blk_sizes, blk_intervals, 
        global_nids, constexprs, keep_zero_params, BLOCK_ID, BLOCK_SIZE
    )

    em_par_update_kernel[grid](
        params, param_flows, cum_pflows, nchs, par_start_ids, pflow_start_ids, blk_sizes, blk_intervals,
        global_nids, constexprs, keep_zero_params, BLOCK_ID, BLOCK_SIZE
    )

    return None


def sgd_par_update(params: torch.Tensor, param_grads: torch.Tensor, par_update_kwargs: Sequence, 
                   lr: float, keep_zero_params: bool = True):
    """
    Apply one-step SGD parameter update.

    :param params: the parameter tensor
    :type params: torch.Tensor

    :param param_grads: gradients of the log-parameters
    :type param_grads: torch.Tensor

    :param lr: learning rate
    :type lr: float

    :param keep_zero_params: whether to freeze zero parameters
    :type keep_zero_params: bool
    """

    par_start_ids, pgrad_start_ids, blk_sizes, blk_intervals, global_nids, nchs, cum_pflows, metadata = par_update_kwargs

    tot_num_nodes = metadata["tot_num_nodes"]
    BLOCK_SIZE = metadata["BLOCK_SIZE"]

    num_blocks = par_start_ids.size(0)
    BLOCK_ID = 2048 // BLOCK_SIZE

    grid = (triton.cdiv(num_blocks, BLOCK_ID),)

    constexprs = torch.tensor([lr, num_blocks]).to(params.device)

    sgd_par_update_kernel[grid](
        params, param_grads, par_start_ids, pgrad_start_ids, blk_sizes, blk_intervals,
        global_nids, constexprs, keep_zero_params, BLOCK_ID, BLOCK_SIZE
    )

    return None



import torch
import torch.nn as nn
from typing import Iterator, Any


class FastParamList(nn.ParameterList):

    def __getitem__(self, idx) -> Any:
        return getattr(self, str(idx))

    def __iter__(self) -> Iterator[Any]:
        return iter(self[i] for i in range(len(self)))



import numpy as np
import torch
import torch.nn as nn
from torch.optim.lr_scheduler import LRScheduler
from typing import Dict, Optional, List

from .optim import CircuitOptimizer


class CircuitScheduler():

    SUPPORTED_SCHEDULER_METHODS = ["constant", "multi_linear"]

    def __init__(self, optimizer: CircuitOptimizer, base_scheduler: Optional[LRScheduler] = None, 
                 method: str = "constant", **kwargs):

        self.optimizer = optimizer
        
        self.base_scheduler = base_scheduler

        assert method in self.SUPPORTED_SCHEDULER_METHODS, f"Unsupported optimization method {method} for circuits."
        self.method = method

        for key, value in kwargs.items():
            self.__dict__[key] = value

        self.step_count = 0

        if self.method == "constant":
            self.lr = self.optimizer.lr
        elif self.method == "multi_linear":
            assert "lrs" in self.__dict__
            assert "milestone_steps" in self.__dict__

            if isinstance(self.lrs, List):
                self.lrs = np.array(self.lrs)
            else:
                try:
                    self.lrs = np.array(list(self.lrs))
                except Exceptions:
                    pass

            if isinstance(self.milestone_steps, List):
                self.milestone_steps = np.array(self.milestone_steps)
            else:
                try:
                    self.milestone_steps = np.array(list(self.milestone_steps))
                except Exceptions:
                    pass
            
            assert isinstance(self.lrs, np.ndarray)
            assert isinstance(self.milestone_steps, np.ndarray)
        else:
            raise ValueError(f"Unknown method {self.method}.")

    def step(self):
        if self.base_scheduler is not None:
            self.base_scheduler.step()

        if self.method == "constant":
            self.optimizer.lr = self.lr
        elif self.method == "multi_linear":
            if self.step_count >= self.milestone_steps[-1]:
                self.optimizer.lr = self.lrs[-1]
            else:
                idx = np.sum(self.milestone_steps < self.step_count)
                if idx == 0:
                    self.optimizer.lr = self.lrs[0]
                else:
                    self.optimizer.lr = self.lrs[idx-1] + (self.lrs[idx] - self.lrs[idx-1]) * \
                        (self.step_count - self.milestone_steps[idx-1]) / \
                        (self.milestone_steps[idx] - self.milestone_steps[idx-1])

        self.step_count += 1

    def state_dict(self):
        if self.base_optimizer is not None:
            state_dict = self.base_optimizer.state_dict()
        else:
            state_dict = dict()

        state_dict["pc_states"] = dict()
        for key, value in self.__dict__.items():
            if key != "optimizer" and key != "base_scheduler":
                state_dict["pc_states"][key] = value

    def load_state_dict(self, state_dict: Dict):

        for key, value in state_dict["circuit_states"].items():
            self.__dict__[key] = value

        del state_dict["pc_states"]

        if self.base_optimizer is not None:
            self.base_optimizer.load_state_dict(state_dict)


from __future__ import annotations

import torch
import networkx as nx
from typing import Type, Optional
from pyjuice.graph import *
from pyjuice.layer import *
from pyjuice.model import TensorCircuit
from pyjuice.nodes import multiply, summate, inputs, set_block_size, CircuitNodes
from pyjuice.nodes.distributions import Distribution, Categorical
from pyjuice.utils.util import max_cdf_power_of_2
from .compilation import BayesianTreeToHiddenRegionGraph


def HMM(seq_length: int, num_latents: int, num_emits: int, homogeneous: bool = True, block_size: Optional[int] = None,
        alpha: Optional[torch.Tensor] = None, beta: Optional[torch.Tensor] = None, gamma: Optional[torch.Tensor] = None):
    """
    Constructs Hidden Markov Models.

    :param seq_length: sequence length
    :type seq_length: int

    :param num_latents: size of the latent space
    :type num_latents: int

    :param num_emits: size of the emission space
    :type num_emits: int

    :param homogeneous: whether to define a homogeneous (or inhomogeneous) HMM
    :type homogeneous: bool

    :param block_size: block size of the PC
    :type block_size: Optional[int]

    :param alpha: optional transition parameters of size `[num_latents, num_latents]`
    :type alpha: Optional[torch.Tensor]

    :param beta: optional emission parameters of size `[num_latents, num_emits]`
    :type beta: Optional[torch.Tensor]

    :param gamma: optional init parameters of size `[num_latents]`
    :type gamma: Optional[torch.Tensor]
    """
    
    if block_size is None:
        block_size = min(max_cdf_power_of_2(num_latents), 1024)
    num_node_blocks = num_latents // block_size
    
    with set_block_size(block_size = block_size):

        ns_input = inputs(
            seq_length - 1, num_node_blocks = num_node_blocks,
            dist = Categorical(num_cats = num_emits)
        )

        if beta is not None:
            assert beta.size(0) == num_latents and beta.size(1) == num_emits
            ns_input.set_params(beta)
        
        ns_sum = None
        curr_zs = ns_input
        for var in range(seq_length - 2, -1, -1):
            curr_xs = ns_input.duplicate(var, tie_params = homogeneous)
            
            if ns_sum is None:
                ns = summate(curr_zs, num_node_blocks = num_node_blocks)
                if alpha is not None:
                    assert alpha.size(0) == num_latents and alpha.size(1) == num_latents
                    ns.set_params(alpha)
                ns_sum = ns
            else:
                ns = ns_sum.duplicate(curr_zs, tie_params = homogeneous)

            curr_zs = multiply(curr_xs, ns)
            
        ns = summate(curr_zs, num_node_blocks = 1, block_size = 1)

        if gamma is not None:
            assert gamma.dim() == 1 and gamma.size(0) == num_latents
            ns.set_params(gamma.unsqueeze(0))
    
    return ns



from __future__ import annotations

import math


def max_cdf_power_of_2(val: int):
    count = 0
    while True:
        halfval = val // 2

        if halfval * 2 != val:
            break

        val = halfval
        count += 1

    return 2 ** count



from __future__ import annotations

import torch
import networkx as nx
from typing import Type, Optional

from pyjuice.nodes import multiply, summate, inputs, set_block_size, CircuitNodes
from pyjuice.nodes.distributions import Distribution
from pyjuice.utils.util import max_cdf_power_of_2


def BayesianTreeToHiddenRegionGraph(tree: nx.Graph, 
                                    root,
                                    num_latents: int,  
                                    InputDist: Type[Distribution], 
                                    dist_params: dict,
                                    num_root_ns: int = 1,
                                    block_size: Optional[int] = None,
                                    tie_input_params: bool = False) -> CircuitNodes:
    """
    Given a Tree Bayesian Network tree T1 (i.e. at most one parents), 
    
        1. Bayesian Network T1 becomes `tree` rooted at `root`
        2. Construct Bayesian bayesian network T2:
            - T2 = copy of T1
            - In T2, convert each variable x_i to a categoriacal latent variable z_i with `num_latents` categories
            - In T2, Adds back x_i nodes, with one edge x_i -> z_i 

        3. Returns a RegionGraph of a probabilistics circuit that is equivalent to T2

    For example, if T1 = "x1 -> x2" , then T2  becomes  x1    x2 
                                                        ^     ^
                                                        |     |
                                                        z1 -> z2   
    """

    # Specify block size
    if block_size is None:
        block_size = min(1024, max_cdf_power_of_2(num_latents))

    num_node_blocks = num_latents // block_size

    # Root the tree at `root`
    clt = nx.bfs_tree(tree, root)
    def children(n: int):
        return [c for c in clt.successors(n)]
    
    # Assert at most one parent
    for n in clt.nodes:
        assert len(list(clt.predecessors(n))) <= 1

    # For input parameter tying
    template_ni = None
    
    def get_input_ns(v):
        nonlocal template_ni
        if tie_input_params:
            if template_ni is None:
                ni = inputs(v, num_node_blocks = num_node_blocks, dist = InputDist(**dist_params))
                template_ni = ni
            else:
                ni = template_ni.duplicate(v, tie_params = True)
        else:
            ni = inputs(v, num_node_blocks = num_node_blocks, dist = InputDist(**dist_params))
        
        return ni

    # Compile the region graph for the circuit equivalent to T2
    node_seq = list(nx.dfs_postorder_nodes(tree, root))
    var2rnode = dict()
    with set_block_size(block_size):
        for v in node_seq:
            chs = children(v)

            if len(chs) == 0:
                # Input Region
                r = get_input_ns(v)
                var2rnode[v] = r
            else:
                # Inner Region
                
                # children(z_v)
                ch_regions = [var2rnode[c] for c in chs]

                # Add x_v to children(z_v)
                leaf_r = get_input_ns(v)
                ch_regions.append(leaf_r)

                rp = multiply(*ch_regions)

                if v == root:
                    if num_root_ns == 1:
                        r = summate(rp, num_node_blocks = num_root_ns, block_size = 1)
                    else:
                        r = summate(rp, num_node_blocks = num_root_ns // block_size, block_size = block_size)
                else:
                    r = summate(rp, num_node_blocks = num_node_blocks)

                var2rnode[v] = r

    root_r = var2rnode[root]
    return root_r